# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: parameters_test
   :platform: Unix
   :synopsis: unittest test classes for plugins

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import sys
import unittest
import numpy as np

import savu.plugins.docstring_parser as dp


class ParametersTest(unittest.TestCase):

    def _get_synopsis(self):
        mod_doc = sys.modules[self.__module__].__doc__
        mod_doc = dp._get_doc_lines(mod_doc)
        return dp._parse_args(mod_doc, '')['synopsis']

    def _set_dict(self, values):
        ddict = {}
        ddict['not_param'] = values[0]
        ddict['hide_param'] = values[1]
        ddict['user_param'] = values[2]
        ddict['param'] = values[3]
        ddict['warn'] = values[4]
        ddict['synopsis'] = values[5]
        ddict['info'] = values[6]
        return ddict

    def _get_parsed_doc(self, function):
        mod_doc_lines = dp._get_doc_lines(sys.modules[self.__module__].__doc__)
        lines = dp._get_doc_lines(function.__doc__)
        return dp._parse_args(mod_doc_lines, lines)

    def test_regular_param(self):
        """
        :param reg_param: Testing a regular param. Default: 3.0.
        """
        parsed_dict = self._get_parsed_doc(self.test_regular_param)
        param = [{'default': 3.0, 'dtype': float, 'name': 'reg_param',
                  'desc': 'Testing a regular param.'}]
        manual_dict = self._set_dict([[], [], [], param, '',
                                     self._get_synopsis(), ''])
        self.assertDictEqual(parsed_dict, manual_dict)

    def test_hidden_param(self):
        """
        :*param hidden_param: Testing a hidden param. Default: None.
        """
        parsed_dict = self._get_parsed_doc(self.test_hidden_param)
        param = [{'default': None, 'dtype': type(None), 'name': 'hidden_param',
                  'desc': 'Testing a hidden param.'}]
        manual_dict = self._set_dict([[], ['hidden_param'], [], param, '',
                                     self._get_synopsis(), ''])
        self.assertDictEqual(parsed_dict, manual_dict)

    def test_user_param(self):
        """
        :u*param user_param: Testing a user param. Default: 10.
        """
        parsed_dict = self._get_parsed_doc(self.test_user_param)
        param = [{'default': 10, 'dtype': int, 'name': 'user_param',
                  'desc': 'Testing a user param.'}]
        manual_dict = self._set_dict([[], [], ['user_param'], param, '',
                                     self._get_synopsis(), ''])
        self.assertDictEqual(parsed_dict, manual_dict)

    def test_fixed_param(self):
        """
        :*param fixed_param: Testing a fixed param. Default: ['s1', 's2'].
        """
        parsed_dict = self._get_parsed_doc(self.test_fixed_param)
        param = [{'default': ['s1', 's2'], 'dtype': list,
                  'name': 'fixed_param', 'desc': 'Testing a fixed param.'}]
        manual_dict = self._set_dict([[], ['fixed_param'], [], param, '',
                                     self._get_synopsis(), ''])
        self.assertDictEqual(parsed_dict, manual_dict)

    def test_removed_param(self):
        """
        :~param removed_param: Testing a removed param. Default: {}.
        """
        parsed_dict = self._get_parsed_doc(self.test_removed_param)
        manual_dict = self._set_dict([['removed_param'], [], [], [], '',
                                     self._get_synopsis(), ''])
        self.assertDictEqual(parsed_dict, manual_dict)


if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: XrdAbsorptionApproximation
   :platform: Unix
   :synopsis: A plugin apply xrd absorption approximation using stxm data
.. moduleauthor:: Stephen W.T. Price <stephen.price@diamond.ac.uk>
"""

import numpy as np
import logging
from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
import _xraylib as xl


@register_plugin
class XrdAbsorptionApproximation(BaseFilter, CpuPlugin):
    """
    McNears absorption correction, takes in a normalised absorption sinogram \
    and xrd sinogram stack

    a base absorption correction for stxm and xrd
    :param azimuthal_offset: angle between detectors. Default: 0.
    :param density: the density. Default: 3.5377.
    :param compound: the compount. Default: 'Co0.2(Al2O3)0.8'.
    :param log_me: should we log the transmission. Default: 1.
    """

    def __init__(self):
        logging.debug("Starting Xrd Absorption Approximation")
        super(XrdAbsorptionApproximation,
              self).__init__("XrdAbsorptionApproximation")

    def pre_process(self):
        compound = self.parameters['compound']
        density = self.parameters['density']
        mData = self.get_in_meta_data()[0]
        mono_energy = mData.get('mono_energy')
        peak_energy = mData.get('mono_energy')
        pump_mu = self.get_mu(compound, float(mono_energy), density)
        # sets outgoing x-ray energy to incoming x-ray energy
        peak_mu = self.get_mu(compound, float(mono_energy), density)
        logging.debug("THE PUMP MU IS is:"+str(pump_mu)+str(mono_energy))
        logging.debug("THE PEAK MU IS is:"+str(peak_mu)+str(peak_energy))

        self.atten_ratio = peak_mu/pump_mu
        theta = mData.get('rotation_angle')
        self.dtheta = theta[1]-theta[0]
        if np.abs(self.dtheta) > 10.0:
            logging.warning('Theta step is greater than 10 degrees! Watch out!')
        self.npix_displacement = \
            self.parameters['azimuthal_offset']//self.dtheta
        logging.debug('This gives a pixel offset of %s'
                      % str(self.npix_displacement))
        if self.parameters['log_me']:
            self.stxm = np.log10(np.squeeze(
                    self.exp.index['in_data']['stxm'].data[...]))
        self.stxm = np.squeeze(self.exp.index['in_data']['stxm'].data[...])

    def process_frames(self, data):
        xrd = data[0]
        stxm_orig = self.stxm
        logging.debug('the xrd shape is %s' % str(xrd.shape))
        logging.debug('the stxm shape is %s' % str(stxm_orig.shape))
        stxm = stxm_orig
        absorption = np.roll(stxm, int(self.npix_displacement), axis=0)
        corrected_xrd, corr_fac = self.correct_sino(1., xrd, absorption)
        return corrected_xrd

    def correct_sino(self, Ti_ratio, FFI0_Ti, absorption):
        trans_ave_array = np.ndarray(shape=FFI0_Ti.shape, dtype=np.float64)
        for n in range(len(absorption)):  # columns, axis=0
            row = absorption[n]
            trans_ave_array[n] = row
            for m in range(1, len(row), 1):  # row, axis=1
                # average of all absorption between pixel 0 and pixel
                # - removes t so should be mu only
                trans_ave_array[n][m] = np.average(row[m:-1])

        trans_ave_array = np.nan_to_num(trans_ave_array)
        exponent_Ti = \
            self.get_exponent_Ti_mu(Ti_ratio, absorption, trans_ave_array)

        FFI0_corrected_Ti = np.multiply(FFI0_Ti, exponent_Ti)
        return FFI0_corrected_Ti, exponent_Ti

    def get_exponent_Ti_mu(self, Ti_ratio, absorption, trans_ave_array):
        FF_Ti_xray_mu = np.multiply(trans_ave_array, Ti_ratio)
        exponent_Ti = np.add(FF_Ti_xray_mu, absorption)
        exponent_Ti = np.multiply(exponent_Ti, 1)
        if not self.parameters['log_me']:
            exponent_Ti = np.exp(exponent_Ti)
        return exponent_Ti

    def setup(self):
        logging.debug('setting up the base absorption correction')
        in_dataset, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_meta_data = in_dataset[0].meta_data
        idx = [in_meta_data.get("mono_energy")]
        self.nChannels = len(idx)
        in_pData[0].plugin_data_setup('SINOGRAM', self.get_num_channels())
        spectra = out_datasets[0]
        spectra.create_dataset(in_dataset[0])

        out_pData[0].plugin_data_setup('SINOGRAM', self.get_num_channels())

    def get_num_channels(self):
        return self.nChannels

    def get_azimuthal_offset(self):
        return self.parameters['azimuthal_offset']

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'single'

    def get_mu(self, compound, energy, density):
        '''
        returns mu for a compound for a single energy
        '''
        return (xl.CS_Total_CP(compound, energy)*density)*1e-2



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: 2D segmentation using Morphological Level Sets or active countours
   :platform: Unix
   :synopsis: Wrapper around Morphsnakes for Morphological countours segmentation

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np
# using Morphological snakes module from
# https://github.com/pmneila/morphsnakes
from morphsnakes import morphological_chan_vese

@register_plugin
class MorphSnakes(Plugin, CpuPlugin):
    """
    A Plugin to segment reconstructed data using Morphsnakes module. When initialised
    with a mask, the active contour propagates to find the minimum of energy (a possible edge countour).

    :param lambda1:  Weight parameter for the outer region, if lambda1 is larger than \
        lambda2, the outer region will contain a larger range of values than \
        the inner region. Default: 1.
    :param lambda2:  Weight parameter for the inner region, if lambda2 is larger than \
        lambda1, the inner region will contain a larger range of values than \
        the outer region. Default: 1.
    :param smoothing: Number of times the smoothing operator is applied per iteration, \
        reasonable values are around 1-4 and larger values lead to smoother segmentations. Default: 1.
    :param iterations: The number of iterations. Default: 350.
    :param pattern: pattern to apply this to. Default: "VOLUME_YZ".
    :param out_datasets: The default names. Default: ['MASK_MORPH_EVOLVED'].
    """

    def __init__(self):
        super(MorphSnakes, self).__init__("MorphSnakes")

    def setup(self):

        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')
        in_pData[1].plugin_data_setup(self.parameters['pattern'], 'single') # the initialised mask

        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        out_pData[0].plugin_data_setup(self.parameters['pattern'], 'single')

    def pre_process(self):
        # extract given parameters
        self.lambda1 = self.parameters['lambda1']
        self.lambda2 = self.parameters['lambda2']
        self.smoothing = self.parameters['smoothing']
        self.iterations = self.parameters['iterations']

    def process_frames(self, data):
        # run MorphSnakes here:
        if (np.sum(data[1]) > 0):
            segment_result = morphological_chan_vese(data[0], iterations=self.iterations, lambda1=self.lambda1, lambda2=self.lambda2, init_level_set=data[1])
        else:
            segment_result = np.uint8(np.zeros(np.shape(data[0])))
        return [segment_result]

    def nInput_datasets(self):
        return 2
    def nOutput_datasets(self):
        return 1



'''
run_savu
This is a refactor of the code that used to be contained in dawn.
It's used to mock up a runner for individual savu plugins from a python shell.
It is currently very early in development and will be subject to massive refactor in the future.
'''
from savu.data.experiment_collection import Experiment
from savu.data.meta_data import MetaData
from savu.plugins.utils import get_plugin
import savu.plugins.loaders.utils.yaml_utils as yaml
import os, sys
import numpy as np
from copy import deepcopy as copy
import time
from collections import OrderedDict


def get_output_rank(path2plugin, inputs, params, persistence):
    sys_path_0_lock = persistence['sys_path_0_lock']
    sys_path_0_lock.acquire()
    try:
        parameters = {}
        # slight repack here
        for key in list(params.keys()):
            val = params[key]["value"]
            if type(val)==type(''):
                val = val.replace('\n','').strip()
                parameters[key] = val
        plugin = _savu_setup(path2plugin, inputs, parameters)
        persistence['plugin_object'] = plugin
    finally:
        sys_path_0_lock.release()

    return len(plugin.get_plugin_out_datasets()[0].get_core_dimensions())


def runSavu(path2plugin, params, metaOnly, inputs, persistence):
    '''
    path2plugin  - is the path to the user script that should be run
    params - are the savu parameters
    metaOnly - a boolean for whether the data is kept in metadata or is passed as data
    inputs      - is a dictionary of input objects 
    '''
    t1 = time.time()
    sys_path_0_lock = persistence['sys_path_0_lock']
    sys_path_0_set = persistence['sys_path_0_set']
    plugin_object = persistence['plugin_object']
    axis_labels = persistence['axis_labels']
    axis_values = persistence['axis_values']
    string_key = persistence['string_key']
    parameters = persistence['parameters']
    aux = persistence['aux']
    sys_path_0_lock.acquire()
    try:
        result = copy(inputs)

        scriptDir = os.path.dirname(path2plugin)
        sys_path_0 = sys.path[0]
        if sys_path_0_set and scriptDir != sys_path_0:
            raise Exception("runSavu attempted to change sys.path[0] in a way that "
                            "could cause a race condition. Current sys.path[0] is {!r}, "
                            "trying to set to {!r}".format(sys_path_0, scriptDir))
        else:
            sys.path[0] = scriptDir
            sys_path_0_set = True
        
        if not plugin_object:
            parameters = {}
                # slight repack here
            for key in list(params.keys()):
#                 print "here"
                val = params[key]["value"]
                if type(val)==type(''):
                    val = val.replace('\n','').strip()
#                 print val
                parameters[key] = val
                print(("val: {}".format(val)))
#             print "initialising the object"
            plugin_object = _savu_setup(path2plugin, inputs, parameters)
            persistence['plugin_object'] = plugin_object
            axis_labels, axis_values = process_init(plugin_object)
#             print "I did the initialisation"
#             print "axis labels",axis_labels
#             print "axis_values", axis_values
#             print plugin_object
            chkstring =  [any(isinstance(ix, str) for ix in axis_values[label]) for label in axis_labels]
            if any(chkstring): # are any axis values strings we instead make this an aux out
                metaOnly = True
#                 print "AXIS LABELS"+str(axis_values)
                string_key = axis_labels[chkstring.index(True)]
                aux = OrderedDict.fromkeys(axis_values[string_key])
#                 print aux.keys()
            else:
                string_key = axis_labels[0]# will it always be the first one?
            if not metaOnly:
                if len(axis_labels) == 1:
                    result['xaxis']=axis_values[axis_labels[0]]
                    result['xaxis_title']=axis_labels[0]
                if len(axis_labels) == 2:
#                     print "set the output axes"
                    x = axis_labels[0]
                    result['xaxis_title']=x
                    y = axis_labels[1]
                    result['yaxis_title']=y
                    result['yaxis']=axis_values[y]
                    result['xaxis']=axis_values[x]
        else:
            pass
    finally:
        sys_path_0_lock.release()

    if plugin_object.get_max_frames()>1: # we need to get round this since we are frame independant
        data = np.expand_dims(inputs['data'], 0)
    else:
        data = inputs['data']

    print(("metaOnly: {}".format(metaOnly)))

    if not metaOnly: 

        out = plugin_object.process_frames([data])
#         print "ran the plugin"

        result['data'] = out
    elif metaOnly:
        result['data'] = inputs['data']
#         print type(result['data'])
        out_array = plugin_object.process_frames([inputs['data']])

#         print aux.keys()

        for k,key in enumerate(aux.keys()):
            aux[key]=np.array([out_array[k]])# wow really

        result['auxiliary'] = aux
    t2 = time.time()
    print("time to runSavu = "+str((t2-t1)))
    return result


def _savu_setup(path2plugin, inputs, parameters):
    print("running _savu_setup")
    parameters['in_datasets'] = [inputs['dataset_name']]
    parameters['out_datasets'] = [inputs['dataset_name']]
    plugin = get_plugin(path2plugin.split('.py')[0]+'.py')
    plugin.exp = setup_exp_and_data(inputs, inputs['data'], plugin)
    plugin._set_parameters(parameters)
    plugin._set_plugin_datasets()
    plugin.setup()
    return plugin


def process_init(plugin):
    axis_labels = plugin.get_out_datasets()[0].get_axis_label_keys()
    axis_labels.remove('idx')  # get the labels
    axis_values = {}
    plugin._clean_up()  # this copies the metadata!
    for label in axis_labels:
        axis_values[label] = plugin.get_out_datasets()[0].meta_data.get(label)
    plugin.base_pre_process()
    plugin.pre_process()
    return axis_labels, axis_values


def setup_exp_and_data(inputs, data, plugin):
    exp = DawnExperiment(get_options())
    data_obj = exp.create_data_object('in_data', inputs['dataset_name'])
    data_obj.data = None
    if len(inputs['data'].shape)==1:
#         print data.shape
        if inputs['xaxis_title'] is None or inputs['xaxis_title'].isspace():
            inputs['xaxis_title']='x'
            inputs['xaxis'] = np.arange(inputs['data'].shape[0])
        data_obj.set_axis_labels('idx.units', inputs['xaxis_title'] + '.units')
        data_obj.meta_data.set('idx', np.array([1]))
        data_obj.meta_data.set(str(inputs['xaxis_title']), inputs['xaxis'])
        data_obj.add_pattern(plugin.get_plugin_pattern(), core_dims=(1,), slice_dims=(0, ))
        data_obj.add_pattern('SINOGRAM', core_dims=(1,), slice_dims=(0, )) # good to add these two on too
        data_obj.add_pattern('PROJECTION', core_dims=(1,), slice_dims=(0, ))
    if len(inputs['data'].shape)==2:
        if inputs['xaxis_title'] is None  or inputs['xaxis_title'].isspace():
            print("set x")
            inputs['xaxis_title']='x'
            inputs['xaxis'] = np.arange(inputs['data'].shape[0])
        if inputs['yaxis_title'] is None or inputs['yaxis_title'].isspace():
            print("set y")
            inputs['yaxis_title']='y'
            size_y_axis = inputs['data'].shape[1]
            inputs['yaxis'] = np.arange(size_y_axis)
        
        data_obj.set_axis_labels('idx.units', inputs['xaxis_title'] + '.units', inputs['yaxis_title'] + '.units')
        data_obj.meta_data.set('idx', np.array([1]))
        data_obj.meta_data.set(str(inputs['xaxis_title']), inputs['xaxis'])
        data_obj.meta_data.set(str(inputs['yaxis_title']), inputs['yaxis'])
        data_obj.add_pattern(plugin.get_plugin_pattern(), core_dims=(1,2,), slice_dims=(0, ))
        data_obj.add_pattern('SINOGRAM', core_dims=(1,2,), slice_dims=(0, )) # good to add these two on too
        data_obj.add_pattern('PROJECTION', core_dims=(1,2,), slice_dims=(0, ))
   
    data_obj.set_shape((1, ) + data.shape) # need to add for now for slicing...
    data_obj.get_preview().set_preview([])
    return exp

class DawnExperiment(Experiment):
    def __init__(self, options):
        self.index={"in_data": {}, "out_data": {}, "mapping": {}}
        self.meta_data = MetaData(get_options())
        self.nxs_file = None

def get_options():
    options = {}
    options['dawn_runner'] = True
    options['transport'] = 'hdf5'
    options['process_names'] = 'CPU0'
    options['processes'] = 'CPU0'
    options['data_file'] = ''
    options['process_file'] = ''
    options['out_path'] = ''
    options['inter_path'] = ''
    options['log_path'] = ''
    options['run_type'] = ''
    options['verbose'] = 'True'
    options['system_params'] = _set_system_params()
    return options

def _set_system_params():
    # look in conda environment to see which version is being used
    savu_path = sys.modules['savu'].__path__[0]
    sys_files = os.path.join(
            os.path.dirname(savu_path), 'system_files')
    subdirs = os.listdir(sys_files)
    sys_folder = 'dls' if len(subdirs) > 1 else subdirs[0]
    fname = 'system_parameters.yml'
    sys_file = os.path.join(sys_files, sys_folder, fname)
    return yaml.read_yaml(sys_file)



'''
Created on 24 May 2016

@author: ssg37927
'''

import sys
import unittest
from mock import patch
from io import StringIO

from scripts.config_generator import savu_config


class Test(unittest.TestCase):
    def savu_config_runner(self, input_list, output_checks):
        with patch('builtins.input', side_effect=input_list):

            saved_stdout = sys.stdout
            try:
                out = StringIO()
                sys.stdout = out
                savu_config.main(test=True)
                output = out.getvalue().strip()
                for check in output_checks:
                    assert check in output
            finally:
                sys.stdout = saved_stdout

    def testExit(self):
        input_list = ['exit', 'y']
        output_checks = ['Thanks for using the application']
        self.savu_config_runner(input_list, output_checks)

    def testHelpBlank(self):
        input_list = ['', 'exit', 'y']
        output_checks = [
            'help :  Display the help information', 'exit :  Close the program'
        ]
        self.savu_config_runner(input_list, output_checks)

    def testHelpCommand(self):
        input_list = ['help', 'exit', 'y']
        output_checks = [
            'help :  Display the help information', 'exit :  Close the program'
        ]
        self.savu_config_runner(input_list, output_checks)


if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: image_data
   :platform: Unix
   :synopsis: A module for loading any of the FabIO python module supported \
       image formats (e.g. tiffs)

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import fabio
import numpy as np

from savu.data.data_structures.data_types.base_type import BaseType


class ImageData(BaseType):
    """ This class loads any of the FabIO python module
    supported image formats. """

    def __init__(self, folder, Data, dim, shape=None, data_prefix=None):
        self.folder = folder
        self._data_obj = Data
        self.frame_dim = dim
        self.shape = shape
        self.prefix = data_prefix
        super(ImageData, self).__init__()

        self.nFrames = None

        self.file_names = self.__get_file_names(folder, data_prefix)
        self.start_file = fabio.open(self.file_names[0])
        self.dtype = self.start_file.data[0, 0].dtype
        self.image_shape = (self.start_file.dim2, self.start_file.dim1)

        if shape is None:
            self.shape = (self.nFrames,)
        else:
            self.shape = shape

        self.full_shape = self.image_shape + self.shape
        self.image_dims = set(np.arange(len(self.full_shape)))\
            .difference(set(self.frame_dim))

    def clone_data_args(self, args, kwargs, extras):
        args = ['folder', 'self', 'frame_dim']
        kwargs['shape'] = 'shape'
        kwargs['prefix'] = 'prefix'
        return args, kwargs, extras

    def __getitem__(self, index):
        index = [index[i] if index[i].start is not None else
                 slice(0, self.shape[i]) for i in range(len(index))]
        size = [len(np.arange(i.start, i.stop, i.step)) for i in index]
        data = np.empty(size, dtype=self.dtype)
        tiff_slices = [index[i] for i in self.image_dims]

        # shift tiff dims to start from 0
        index = list(index)
        for i in self.image_dims:
            end = \
                len(np.arange(0, index[i].stop-index[i].start, index[i].step))
            index[i] = slice(0, end, 1)

        index, frameidx = self.__get_indices(index, size)

        for i in range(len(frameidx)):
            image = fabio.open(self.file_names[frameidx[i]]).data[tuple(tiff_slices)]
            for d in self.frame_dim:
                image = np.expand_dims(image, axis=d)
            data[tuple(index[i])] = image

        return data

    def __get_file_names(self, folder, prefix):
        import re
        import glob
#        files = os.listdir(folder)
        fullpath = str.strip(folder)
        if prefix is not None:
            fullpath = os.path.join(folder, prefix + '*')
        else:
            fullpath = os.path.join(fullpath, '*')
        files = glob.glob(fullpath)
        self.nFrames = len(files)
        file_nos = [int(re.findall(r'\d+', f)[-1]) for f in files]
        sort_idx = np.argsort(file_nos)
        self.start_no = file_nos[sort_idx[0]]
        return list(np.array(files)[sort_idx])

    def get_shape(self):
        dims = list(self.image_dims) + self.frame_dim
        shape = [x for _, x in sorted(zip(dims, self.full_shape))]
        return tuple(shape)

    def __get_idx(self, dim, sl, shape):
        c = int(np.prod(shape[0:dim]))
        r = int(np.prod(shape[dim+1:]))
        vals = np.arange(sl.start, sl.stop, sl.step)
        vals_shift = np.arange(0, len(vals), 1)
        return [np.ravel(np.kron(v, np.ones((r, c)))) for v in \
                [vals, vals_shift]]

    def __get_indices(self, index, size):
        """ Get the indices for the new data array and the file numbers. """
        # indices for non-image dims only
        sub_idx = np.array(index)[np.array(self.frame_dim)]
        sub_size = [size[i] for i in self.frame_dim]

        idx_list = []
        frame_list = []
        for dim in range(len(sub_idx)):
            frame, idx = self.__get_idx(dim, sub_idx[dim], sub_size)
            frame_list.append(frame.astype(int))
            idx_list.append(idx.astype(int))

        lshape = idx_list[0].shape[0]  # this is just size of first frame dim? sub_size[0]
        index = np.tile(index, (lshape, 1))
        frameidx = np.zeros(lshape)

        for dim in range(len(sub_idx)):
            index[:, self.frame_dim[dim]] = \
                [slice(i, i+1, 1) for i in idx_list[dim]]
            frameidx[:] += frame_list[dim]*np.prod(self.shape[dim+1:])
        return index.tolist(), frameidx.astype(int)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: hdf5_tomo_saver
   :platform: Unix
   :synopsis: A class to save data to a hdf5 output file.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import logging
import os
import copy

from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils
from savu.plugins.savers.base_saver import BaseSaver
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.chunking import Chunking


@register_plugin
class Hdf5Saver(BaseSaver, CpuPlugin):
    """
    A class to save tomography data to a hdf5 file

    :param pattern: Optimise data storage to this access pattern: 'optimum' \
        will automate this process by choosing the output pattern from the \
        previous plugin, if it exists, else the first \
        pattern. Default: 'optimum'.
    """

    def __init__(self, name='Hdf5Saver'):
        super(Hdf5Saver, self).__init__(name)
        self.in_data = None
        self.out_data = None
        self.data_name = None
        self.filename = None
        self.group_name = None

    def pre_process(self):
        # Create the hdf5 output file
        self.hdf5 = Hdf5Utils(self.exp)
        self.in_data = self.get_in_datasets()[0]
        self.data_name = self.in_data.get_name()
        current_pattern = self.__set_current_pattern()
        pattern_idx = {'current': current_pattern, 'next': []}

        self.filename = self.__get_file_name()
        self.group_name = self._get_group_name(self.data_name)
        logging.debug("creating the backing file %s", self.filename)
        self.backing_file = self.hdf5._open_backing_h5(self.filename, 'w')
        group = self.backing_file.create_group(self.group_name)
        group.attrs['NX_class'] = 'NXdata'
        group.attrs['signal'] = 'data'
        self.exp._barrier()
        shape = self.in_data.get_shape()
        chunking = Chunking(self.exp, pattern_idx)
        dtype = self.in_data.data.dtype
        chunks = chunking._calculate_chunking(shape, dtype)
        self.exp._barrier()
        self.out_data = self.hdf5.create_dataset_nofill(
                group, "data", shape, dtype, chunks=chunks)

    def process_frames(self, data):
        self.out_data[self.get_current_slice_list()[0]] = data[0]

    def post_process(self):
        self._link_datafile_to_nexus_file(self.data_name, self.filename,
                                          self.group_name + '/data')
        self.backing_file.close()

    def __set_current_pattern(self):
        pattern = copy.deepcopy(self.in_data._get_plugin_data().get_pattern())
        pattern[list(pattern.keys())[0]]['max_frames'] = self.get_max_frames()
        return pattern

    def get_pattern(self):
        if self.parameters['pattern'] != 'optimum':
            return self.parameters['pattern']
        previous_pattern = self.get_in_datasets()[0].get_previous_pattern()
        if previous_pattern:
            return list(previous_pattern.keys())[0]
        return list(self.get_in_datasets()[0].get_data_patterns().keys())[0]

    def __get_file_name(self):
        nPlugin = self.exp.meta_data.get('nPlugin')
        plugin_dict = \
            self.exp._get_experiment_collection()['plugin_dict'][nPlugin]
        fname = self.data_name + '_p' + str(nPlugin) + '_' + \
            plugin_dict['id'].split('.')[-1] + '.h5'
        out_path = self.exp.meta_data.get('out_path')
        return os.path.join(out_path, fname)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: slice_lists
   :platform: Unix
   :synopsis: Contains classes for creating global and local slice lists

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np


class SliceLists(object):
    """
    SliceLists class creates global and local slices lists used to transfer
    the data
    """

    def __init__(self, name='SliceLists'):
        super(SliceLists, self).__init__()
        self.pad = False
        self.transfer_data = None

    def _get_process_data(self):
        return self.process_data

    def _single_slice_list(self, nSlices, nDims, core_slice, core_dirs,
                           slice_dirs, fix, index):

        fix_dirs, value = fix
        slice_list = []
        for i in range(nSlices):
            getitem = np.array([slice(None)]*nDims)
            getitem[core_dirs] = core_slice[np.arange(len(core_dirs))]
            for f in range(len(fix_dirs)):
                getitem[fix_dirs[f]] = slice(value[f], value[f] + 1, 1)
            for sdir in range(len(slice_dirs)):
                getitem[slice_dirs[sdir]] = slice(index[sdir, i],
                                                  index[sdir, i] + 1, 1)
            slice_list.append(tuple(getitem))
        return slice_list

    def _get_slice_dirs_index(self, slice_dirs, shape, value, calc=None):
        """
        returns a list of arrays for each slice dimension, where each array
        gives the indices for that slice dimension.
        """
        # create the indexing array
        chunk, length, repeat = self.__chunk_length_repeat(slice_dirs, shape)
        values = None
        idx_list = []
        for i in range(len(slice_dirs)):
            c = chunk[i]
            r = repeat[i]
            values = eval(value)
            idx = np.ravel(np.kron(values, np.ones((r, c))))
            idx_list.append(idx.astype(int))
        return np.array(idx_list)

    def __chunk_length_repeat(self, slice_dirs, shape):
        """
        For each slice dimension, determine 3 values relevant to the slicing.

        :returns: chunk, length, repeat
            chunk: how many repeats of the same index value before an increment
            length: the slice dimension length (sequence length)
            repeat: how many times does the sequence of chunked numbers repeat
        :rtype: [int, int, int]
        """
        sshape = self.__get_shape_of_slice_dirs(slice_dirs, shape)
        if not slice_dirs:
            return [1], [1], [1]

        chunk = []
        length = []
        repeat = []
        for dim in range(len(slice_dirs)):
            chunk.append(int(np.prod(sshape[0:dim])))
            length.append(sshape[dim])
            repeat.append(int(np.prod(sshape[dim+1:])))
        return chunk, length, repeat

    def __get_shape_of_slice_dirs(self, slice_dirs, shape):
        sshape = [shape[sslice] for sslice in slice_dirs]
        if 'var' in sshape:
            shape = list(shape)
            for index, value in enumerate(shape):
                if isinstance(value, str):
                    shape[index] = \
                        len(self.data_info.get('axis_labels')[index])
            shape = tuple(shape)
            sshape = [shape[sslice] for sslice in slice_dirs]
        return sshape

    def _get_core_slices(self, core_dirs):
        core_slice = []
        starts, stops, steps, chunks = \
            self.data.get_preview().get_starts_stops_steps()

        for c in core_dirs:
            if (chunks[c]) > 1:
                if (stops[c] - starts[c] == 1):
                    start = starts[c] - int(chunks[c] / 2.0)
                    if start < 0:
                        raise Exception('Cannot have a negative value in the '
                                        'slice list.')
                    stop = starts[c] + (chunks[c] - int(chunks[c] / 2.0))
                    core_slice.append(slice(start, stop, 1))
                else:
                    raise Exception("The core dimension does not support "
                                    "multiple chunks.")
            else:
                core_slice.append(slice(starts[c], stops[c], steps[c]))
        return np.array(core_slice)

    def _banked_list(self, slice_list, max_frames, pad=False):
        shape = self.data.get_shape()
        slice_dirs = self.data.get_slice_dimensions()
        chunk, length, repeat = self.__chunk_length_repeat(slice_dirs, shape)
        sdir_shape = [shape[i] for i in slice_dirs]
        split, split_dim = self.__get_split_length(max_frames, sdir_shape)
        split_list = self._split_list(slice_list, split)

        banked = []
        for s in split_list:
            b = self._split_list(s, max_frames)
            if pad:
                diff = max_frames - len(b[-1])
                #print("diff", diff, max_frames, len(b[-1]))
                b[-1][-1] = self._fix_list_length(b[-1][-1], diff, split_dim) if diff \
                    else b[-1][-1]
            banked.extend(b)
        return banked

    def __get_split_length(self, max_frames, shape):
        nDims = 0
        while(nDims < len(shape)):
            nDims += 1
            prod = np.prod([shape[i] for i in range(nDims)])
            if prod/float(max_frames) >= 1:
                break
        return prod, nDims-1

    def _group_dimension(self, sl, dim, step):
        start = sl[0][dim].start
        stop = sl[-1][dim].stop
        working_slice = list(sl[0])
        working_slice[dim] = slice(start, stop, step)
        return tuple(working_slice)

    def _split_list(self, the_list, size):
        return [the_list[x:x+size] for x in range(0, len(the_list), size)]

    # This method only works if the split dimensions in the slice list contain
    # slice objects
    def __split_frames(self, slice_list, split_list):
        split = [list(map(int, a.split('.'))) for a in split_list]
        dims = [s[0] for s in split]
        length = [s[1] for s in split]
        replace = self.__get_split_frame_entries(slice_list, dims, length)
        # now replace each slice list entry with multiple entries
        array_list = []
        for sl in slice_list:
            new_list = np.array([sl for i in range(len(replace[0]))])
            for d, i in zip(dims, list(range(len(dims)))):
                new_list[:, d] = replace[i]
            array_list += [tuple(a) for a in new_list]

        return tuple(array_list)

    def __get_split_frame_entries(self, slice_list, dims, length):
        shape = self.get_shape
        replace = []
        seq_len = []

        # get the new entries
        for d, l in zip(dims, length):
            sl = slice_list[0][d]
            start = 0 if sl.start is None else sl.start
            stop = shape[d] if sl.stop is None else sl.stop
            inc = l*sl.step if sl.step else l
            temp_list = [slice(a, a+inc) for a in np.arange(start, stop, inc)]
            if temp_list[-1].stop > stop:
                temp = temp_list[-1]
                temp_list[-1] = slice(temp.start, stop, temp.step)
            replace.append(temp_list)
            seq_len.append(len(temp_list))

        # calculate the permutations
        length = np.array(seq_len)
        chunk = [int(np.prod(length[0:dim])) for dim in range(len(dims))]
        repeat = [int(np.prod(length[dim+1:])) for dim in range(len(dims))]
        full_replace = []
        for d in range(len(dims)):
            temp = [[replace[d][i]]*chunk[d] for x in range(repeat[d]) for i
                    in range(len(replace[d]))]
            full_replace.append([t for sub in temp for t in sub])
        return full_replace

    def _get_frames_per_process(self, slice_list):
        processes = self.data.exp.meta_data.get("processes")
        process = self.data.exp.meta_data.get("process")
        frame_idx = np.arange(len(slice_list))
        try:
            frames = np.array_split(frame_idx, len(processes))[process]
            slice_list = slice_list[frames[0]:frames[-1]+1]
        except IndexError:
            slice_list = []
        return slice_list, frames

    def _pad_slice_list(self, slice_list, inc_start_str: str, inc_stop_str: str):
        """ Amend the slice lists to include padding.  Includes variations for
        transfer and process slice lists. """
        pData = self.data._get_plugin_data()
        if not pData.padding:
            return slice_list

        pad_dict = pData.padding._get_padding_directions()

        shape = self.data.get_shape()
        for ddir, value in pad_dict.items():
            inc_start = eval(inc_start_str)
            inc_stop = eval(inc_stop_str)
            for i in range(len(slice_list)):
                slice_list[i] = list(slice_list[i])
                sl = slice_list[i][ddir]
                if sl.start is None:
                    sl = slice(0, shape[ddir], 1)
                slice_list[i][ddir] = \
                    slice(sl.start + inc_start, sl.stop + inc_stop, sl.step)
                slice_list[i] = tuple(slice_list[i])
        return slice_list

    def _fix_list_length(self, sl, length, dim):
        sdir = self.data.get_slice_dimensions()
        sl = list(sl)
        s = sl[sdir[dim]]
        sl[sdir[dim]] = slice(s.start, s.stop + length*s.step, s.step)
        return tuple(sl)

    def _get_local_single_slice_list(self, shape):
        slice_dirs = self.data.get_slice_dimensions()
        core_dirs = np.array(self.data.get_core_dimensions())
        fix = [[]]*2
        core_slice = np.array([slice(None)]*len(core_dirs))
        shape = tuple([shape[i] for i in range(len(shape))])
        values = 'np.arange(shape[slice_dirs[i]])'
        index = self._get_slice_dirs_index(slice_dirs, shape, values)
        # there may be no slice dirs
        index = index if index.size else np.array([[0]])
        nSlices = index.shape[1] if index.size else len(fix[0])
        nDims = len(shape)

        ssl = self._single_slice_list(
            nSlices, nDims, core_slice, core_dirs, slice_dirs, fix, index)
        return ssl

    def _group_slice_list_in_one_dimension(self, slice_list, max_frames,
                                           group_dim, pad=False):
        """ Group the slice list in one dimension, stopping at \
        boundaries - prepare a slice list for multi-frame plugin processing.
        """
        if group_dim is None:
            return slice_list

        banked = self._banked_list(slice_list, max_frames, pad=pad)
        grouped = []
        for group in banked:
            sub_groups = self._split_list(group, max_frames)
            for sub in sub_groups:
                grouped.append(self._group_dimension(sub, group_dim, 1))
        return grouped

    def _get_global_single_slice_list(self, shape):
        slice_dirs = self.data.get_slice_dimensions()
        core_dirs = np.array(self.data.get_core_dimensions())
        fix = self.data._get_plugin_data()._get_fixed_dimensions()
        core_slice = self._get_core_slices(core_dirs)
        values = 'self._get_slice_dir_index(slice_dirs[i])'
        index = self._get_slice_dirs_index(slice_dirs, shape, values)
        nSlices = index.shape[1] if index.size else len(fix[0])
        nDims = len(shape)
        ssl = self._single_slice_list(
            nSlices, nDims, core_slice, core_dirs, slice_dirs, fix, index)
        return ssl

    def _group_slice_list_in_multiple_dimensions(self, slice_list, max_frames,
                                                 group_dim, pad=False):
        """ Group the slice list in multiple dimensions - prepare a slice list\
        for file transfer.
        """
        if group_dim is None:
            return slice_list

        steps = self.data.get_preview().get_starts_stops_steps('steps')
        sub_groups = self._banked_list(slice_list, max_frames, pad=pad)

        grouped = []
        for sub in sub_groups:
            temp = list(sub[0])
            for dim in group_dim:
                temp[dim] = self._group_dimension(sub, dim, steps[dim])[dim]
            grouped.append(tuple(temp))

        return grouped

class LocalData(object):
    """ The LocalData class organises the slicing of transferred data to \
    give the shape requested by a plugin for each run of 'process_frames'.
    """

    def __init__(self, dtype, transport_data):
        self.dtype = dtype  # in or out ProcessData object
        self.td = transport_data
        self.data = transport_data.data
        self.pData = self.data._get_plugin_data()
        self.shape = self.data.get_shape()
        self.sdir = None

    def _get_dict(self):
        return self._get_dict_in() if self.dtype == 'in' else \
            self._get_dict_out()

    def _get_dict_in(self):
        sl_dict = {}
        sl = self._get_slice_list()
        sl = self.td._pad_slice_list(sl, '0', 'sum(value.values())')
        sl_dict['process'] = sl
        return sl_dict

    def _get_dict_out(self):
        sl_dict = {}
        sl_dict['process'] = self._get_slice_list()
        sl_dict['unpad'] = self.__get_unpad_slice_list(len(sl_dict['process']))
        return sl_dict

    def _get_slice_list(self):
        """ Splits a file transfer slice list into a list of (padded) slices
        required for each loop of process_frames.
        """
        slice_dirs = self.data.get_slice_dimensions()
        self.sdir = slice_dirs[0] if len(slice_dirs) > 0 else None

        pData = self.pData
        mf_process = pData.meta_data.get('max_frames_process')
        shape = pData.get_shape_transfer()

        process_ssl = self.td._get_local_single_slice_list(shape)

        process_gsl = self.td._group_slice_list_in_one_dimension(
                process_ssl, mf_process, self.sdir, pad=True) # pad if mfp is > nSlices (e.g. mfp = fixed int)
        return process_gsl

    def __get_unpad_slice_list(self, reps):
        # setting process slice list unpad here - not currently working for 4D data
        sl = [slice(None)]*len(self.pData.get_shape_transfer())
        if not self.pData.padding:
            return tuple([tuple(sl)]*reps)
        pad_dict = self.pData.padding._get_padding_directions()
        for ddir, value in pad_dict.items():
            sl[ddir] = slice(value['before'], -value['after'])
        return tuple([tuple(sl)]*reps)


class GlobalData(object):
    """
    The GlobalData class organises the movement and slicing of the data from
    file.
    """

    def __init__(self, dtype, transport):
        self.dtype = dtype  # in or out TransferData object
        self.trans = transport
        self.data = transport.data
        self.pData = self.data._get_plugin_data()
        self.shape = self.data.get_shape()

    def _get_dict(self):
        return self._get_dict_in() if self.dtype == 'in' else \
            self._get_dict_out()

    def _get_dict_in(self):
        sl_dict = {}
        sl, current = \
            self._get_slice_list(self.shape, current_sl=True, pad=True)

        sl_dict['current'], _ = self.trans._get_frames_per_process(current)
        sl, sl_dict['frames'] = self.trans._get_frames_per_process(sl)

        if self.trans.pad:
            sl = self.trans._pad_slice_list(
                sl, "-value['before']", "value['after']")
        sl_dict['transfer'] = sl
        return sl_dict

    def _get_dict_out(self):
        sl_dict = {}
        sl, _ = self._get_slice_list(self.shape)
        sl_dict['transfer'], _ = self.trans._get_frames_per_process(sl)
        return sl_dict

    def _get_slice_list(self, shape, current_sl=None, pad=False):
        mft = self.pData._get_max_frames_transfer()
        transfer_ssl = self.trans._get_global_single_slice_list(shape)

        if transfer_ssl is None:
            raise Exception("Data type %s does not support slicing in "
                            "directions %s" % (self.get_current_pattern_name(),
                                               self.get_slice_directions()))
        slice_dims = self.data.get_slice_dimensions()
        transfer_gsl = self.trans._group_slice_list_in_multiple_dimensions(
                transfer_ssl, mft, slice_dims, pad=pad)

        if current_sl:
            mfp = self.pData._get_max_frames_process()
            current_sl = self.trans._group_slice_list_in_multiple_dimensions(
                    transfer_ssl, mfp, slice_dims, pad=pad)

        split_list = self.pData.split
        transfer_gsl = self.__split_frames(transfer_gsl, split_list) if \
            split_list else transfer_gsl

        return transfer_gsl, current_sl

    def _get_padded_data(self, slice_list, end=False):
        slice_list = list(slice_list)
        pData = self.pData
        pad_dims = list(set(self.data.get_core_dimensions() +
                            (self.data.get_slice_dimensions())))
        pad_list = []
        for i in range(len(slice_list)):
            pad_list.append([0, 0])

        data_dict = self.data.data_info.get_dictionary()
        shape = data_dict['orig_shape'] if 'orig_shape' in list(data_dict.keys()) \
            else self.data.get_shape()

        for dim in range(len(pad_dims)):
            sl = slice_list[dim]
            if sl.start < 0:
                pad_list[dim][0] = -sl.start
                slice_list[dim] = slice(0, sl.stop, sl.step)
            diff = sl.stop - shape[dim]
            if diff > 0:
                pad_list[dim][1] = diff
                slice_list[dim] = \
                    slice(slice_list[dim].start, sl.stop - diff, sl.step)

        data = self.data.data[tuple(slice_list)]

        if np.sum(pad_list):
            mode = pData.padding.mode if pData.padding else 'edge'
            temp = np.pad(data, tuple(pad_list), mode=mode)
            return temp
        return data



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: ptypy_compact
   :platform: Unix
   :synopsis: A plugin to perform ptychography using ptypy

.. moduleauthor:: Aaron Parsons <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.utils import register_plugin
from savu.plugins.ptychography.base_ptycho import BasePtycho
import numpy as np
from ptypy.core import Ptycho
from ptypy import utils as u
from copy import deepcopy as copy
import h5py as h5


@register_plugin
class PtypyBatch(BasePtycho):
    """
    This plugin performs ptychography using the ptypy package. The same parameter set is used across 
    all slices and is based on the output from a previous reconstruction. 
    :param ptyr_file: The ptyd for a previously successful reconstruction. Default: '/dls/science/users/clb02321/DAWN_stable/I13Test_Catalysts/processing/catalyst_data/analysis92713/recons/92713/92713_DM_0030_0.ptyr'.
    :param mask_file: The mask file. Default: '/dls/science/users/clb02321/DAWN_stable/I13Test_Catalysts/processing/catalyst_data/analysis92713/new_mask2.hdf'.
    :param mask_entry: The mask entry. Default: '/mask'.
    """

    def __init__(self):
        super(PtypyBatch, self).__init__("PtypyBatch")

    def pre_process(self):
        b = Ptycho.load_run(self.parameters['ptyr_file'], False) # load in the run but without the data
        p = b.p
        existing_scan = copy(p.scans[list(p.scans.keys())[0]])
        del p.scans[list(p.scans.keys())[0]]
        p.scans.savu = existing_scan
        p.scans.savu.data.source = 'savu'
        p.scans.savu.data.recipe = u.Param()
        if self.parameters['mask_file'] is not None:
            p.scans.savu.data.recipe.mask = h5.File(self.parameters['mask_file'],'r')[self.parameters['mask_entry']][...]
        else:
            p.scans.savu.data.recipe.mask = np.ones(self.in_shape[-2:])
        p.scan.illumination = b.probe.storages['S00G00'].data
        self.p = p

    def filter_frames(self, data):
#         idx = self.get_global_frame_index()# the current frame
        p = self.p
        p.scans.savu.data.recipe.data = data[0]
        positions = self.get_positions()[0]# just assume zero for now
        p.scans.savu.data.recipe.positions = positions.T[:600]
        P = Ptycho(p,level=5)
        object_stack = P.obj.storages['S00G00'].data.astype(np.float)
        probe_stack = P.probe.storages['S00G00'].data.astype(np.float)
        return [probe_stack, object_stack, positions]#] add fourier error, realspace error

    def setup(self):
        in_datasets, out_datasets = self.get_datasets()
        self.in_shape = in_datasets[0].get_shape()
        in_pData, __out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('PROJECTION', self.get_max_frames())
        self.in_slice_dirs= np.array(in_pData[0].get_slice_directions())
        self.in_core_dirs= np.sort(np.array(in_pData[0].get_core_directions()))# sort them to get in the right order
        self.b = Ptycho.load_run(self.parameters['ptyr_file'],False)
        self.p = self.b.p
        BasePtycho.setup(self)

    def get_num_probe_modes(self):
        return self.p.scan.coherence.num_probe_modes
    
    def get_num_object_modes(self):
        return self.p.scan.coherence.num_object_modes

    def set_size_object(self, in_d1, positions, pobj):
        first_core = self.in_core_dirs[0]
        last_core = self.in_core_dirs[-1]
        start_shape = self.in_shape[:first_core]
        self.obj_shape=self.b.obj.storages['S00G00'].data.shape
        #print "object shape is" + str(self.obj_shape)
    
    def set_size_probe(self, probe_shape):
        first_core = self.in_core_dirs[0]
        last_core = self.in_core_dirs[-1]
        start_shape = self.in_shape[:first_core]
        self.probe_size=start_shape+self.b.probe.storages['S00G00'].data.shape
        #print "probe size is" + str(self.probe_size)




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: arg_parsers
   :platform: Unix
   :synopsis: Contains all the argparse instances for configurator \
       input commands.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import argparse


class MyException(Exception):
    def __init__(self, message, args=None):
        super(MyException, self).__init__(message)
        self.args = args

    def args(self):
        return self.args


class ArgumentParser(argparse.ArgumentParser):

    def error(self, message):
        raise MyException(message)

    def exit(self, status=0, message=None):
        raise MyException(message)


def __arg_parser(parser, args, command):
    try:
        args = parser.parse_args(args=args)
    except MyException as e:
        print(e)
        print(f"Please type '{command} -h' for help.")
        args = e.args
    return args


def _config_arg_parser():
    """ Argument parser for command line arguments. """
    desc = "Create process lists of plugins for Savu."
    parser = argparse.ArgumentParser(prog='savu_config.py', description=desc)
    disp_str = "Set the display level to show ALL parameters by default."
    parser.add_argument("-a", "--all", action="store_true", dest="disp_all",
                        help=disp_str, default=False)
    input_str = "Open a Savu process list."
    parser.add_argument("-i", "--input", dest="file", help=input_str)
    parser.add_argument("-e", "--error", dest="error", help="Shows all errors that Savu encounters.",
                        action='store_true', default=False)
    parser.add_argument("--examples", dest="examples", action='store_true',
                        help="Add example plugins", default=False)
    return parser.parse_args()


def __verbosity_arguments(parser, q_str, v_str, vv_str):
    """ Add verbosity arguments to a parser. """
    parser.add_argument("-q", "--quiet", action="store_true", dest="quiet",
                        help="Quiet mode. "+q_str, default=False)
    parser.add_argument("-v", "--verbose", action="store_true", dest="verbose",
                        help="Verbose mode. "+v_str, default=False)
    parser.add_argument("-vv", "--vverbose", action="store_true",
                        dest="vverbose", help="Verbose verbose mode. "+vv_str,
                        default=False)


def _open_arg_parser(args, desc):
    """ Argument parser for open command. """
    parser = ArgumentParser(prog='open', description=desc)
    file_str = "The path to the process list to open."
    parser.add_argument('file', help=file_str)
    parser.add_argument('-s', '--skip', help='Skip broken plugins.',
                        action='store_true', default=False)
    return __arg_parser(parser, args, 'open')


def _disp_arg_parser(args, desc):
    """ Argument parser for disp command. """
    parser = ArgumentParser(prog='disp', description=desc)
    q_str = "Display plugin names only."
    v_str = "Display plugin names, synopsis and parameter details."
    vv_str = \
        "Display plugin names, full synopsis, parameter details and warnings."
    __verbosity_arguments(parser, q_str, v_str, vv_str)

    all_str = "Display ALL parameters (user parameters only by default)."
    parser.add_argument("-a", "--all", action='store_true', help=all_str,
                        default=False)

    parser.add_argument("start", nargs='?', help="Display this list entry.")
    stop_str = "Display entries from start to stop."
    parser.add_argument("stop", nargs='?', help=stop_str)
    return __arg_parser(parser, args, 'disp')


def _list_arg_parser(args, desc):
    """ Argument parser for list command. """
    parser = ArgumentParser(prog='list', description=desc)
    q_str = "List plugin names only."
    v_str = "List plugin names and full synopsis."
    vv_str = "List all information."
    __verbosity_arguments(parser, q_str, v_str, vv_str)
    type_str = "List plugins or collections containing this string."
    parser.add_argument("string", nargs='?', help=type_str, default="")

    return __arg_parser(parser, args, 'list')


def _save_arg_parser(args, desc):
    """ Argument parser for save command. """
    parser = ArgumentParser(prog='save', description=desc)
    parser.add_argument("filepath", nargs='?', help="The output file path.")
    parser.add_argument("-i", "--input", action="store_true", default=False,
                        help="Save to the input file.")
    parser.add_argument("-t", "--template", action="store_true", default=False,
                        help="Create a Savu template file.")
    return __arg_parser(parser, args, 'save')


def _mod_arg_parser(args, desc):
    """ Argument parser for mod command. """
    # required to allow negative values in a list or tuple
    args.insert(1, '--')

    parser = ArgumentParser(prog='mod', description=desc)
    param_str = ("The plugin parameter to modify. Either "
                 "'plugin_pos.param_name' or ' plugin_pos.param_no'")
    parser.add_argument("param", help=param_str)
    val_str = "The plugin parameter value."
    parser.add_argument("value", nargs='+', help=val_str)
    return __arg_parser(parser, args, 'mod')


def _set_arg_parser(args, desc):
    """ Argument parser for set command. """
    parser = ArgumentParser(prog='set', description=desc)
    parser.add_argument('plugin_pos', type=str, nargs='+',
                        help="Plugin position.")
    parser.add_argument("status", type=str, choices=['on', 'ON', 'off', 'OFF'],
                        help="Plugin status.")
    return __arg_parser(parser, args, 'set')


def _add_arg_parser(args, desc):
    """ Argument parser for add command. """
    parser = ArgumentParser(prog='add', description=desc)
    parser.add_argument("name", help="The plugin name.")
    pos_str = "Plugin list position (defaults to end)."
    parser.add_argument('pos', nargs='?', help=pos_str)
    return __arg_parser(parser, args, 'add')


def _ref_arg_parser(args, desc):
    """ Argument parser for ref command. """
    parser = ArgumentParser(prog='ref', description=desc)
    plugin_str = "Plugin position to refresh or '*' for the whole list"
    parser.add_argument("pos", nargs='+', help=plugin_str)
    defaults_str = "Populate parameters with default values."
    parser.add_argument("-d", "--defaults", action="store_true",
                        dest="defaults", help=defaults_str, default=False)
    parser.add_argument("-n", "--nodisp", action="store_true", dest="nodisp",
                        help=argparse.SUPPRESS, default=False)
    return __arg_parser(parser, args, 'ref')


def _rem_arg_parser(args, desc):
    """ Argument parser for rem command. """
    parser = ArgumentParser(prog='rem', description=desc)
    parser.add_argument('pos', help="Plugin position(s).")
    return __arg_parser(parser, args, 'rem')


def _move_arg_parser(args, desc):
    """ Argument parser for move command. """
    parser = ArgumentParser(prog='move', description=desc)
    parser.add_argument("orig_pos", help="Original position.")
    parser.add_argument('new_pos', help="New position.")
    return __arg_parser(parser, args, 'move')


def _coll_arg_parser(args, desc):
    """ Argument parser for coll command. """
    parser = ArgumentParser(prog='coll', description=desc)
    return __arg_parser(parser, args, 'coll')


def _get_verbosity(args):
    if args.vverbose:
        return '-vv'
    elif args.verbose:
        return '-v'
    elif args.quiet:
        return '-q'
    else:
        return None



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Get the Savu version
"""
import os
import pathlib

# for the savu_installer
savu_path = pathlib.Path(__file__).parent.absolute()
__install__ = "install/savu_hpc/savu_installer"
with open(os.path.join(
        os.path.join(savu_path, "..", __install__, "version.txt"))) as f:
    __version__ = f.readline().strip()





# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_driver
   :platform: Unix
   :synopsis: Base class or all driver plugins

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import logging
import numpy as np
from mpi4py import MPI

from savu.plugins.driver.basic_driver import BasicDriver


class PluginDriver(BasicDriver):
    """
    The base class from which all plugins should inherit.
    """

    def __init__(self):
        super(PluginDriver, self).__init__()
        self._communicator = None

    def _run_plugin_instances(self, transport, communicator=MPI.COMM_WORLD):
        """ Runs the pre_process, process and post_process methods.

        If parameter tuning is required, loop over the methods and set the
        correct parameters for each run. """

        self.__set_communicator(communicator)
        out_data = self.get_out_datasets()
        extra_dims = self.extra_dims
        repeat = np.prod(extra_dims) if extra_dims else 1

        param_idx = self.__calc_param_indices(extra_dims)
        out_data_dims = [len(d.get_shape()) for d in out_data]
        param_dims = [list(range(d - len(extra_dims), d)) for d in out_data_dims]

        if extra_dims:
            init_vars = self.__get_local_dict()

        for i in range(repeat):
            if extra_dims:
                self.__reset_local_vars(init_vars)
                self._set_parameters_this_instance(param_idx[i])
                for j in range(len(out_data)):
                    out_data[j]._get_plugin_data()\
                        .set_fixed_dimensions(param_dims[j], param_idx[i])

            super(PluginDriver, self).\
                _run_plugin_instances(transport, communicator=communicator)
            self._reset_process_frames_counter()

        self._revert_preview(self.parameters['in_datasets'])

        for j in range(len(out_data)):
            out_data[j].set_shape(out_data[j].data.shape)

    def __get_local_dict(self):
        """ Gets the local variables of the class minus those from the Plugin
        class. """
        from savu.plugins.plugin import Plugin
        plugin = Plugin()
        copy_keys = vars(self).keys() - vars(plugin).keys()
        copy_dict = {}
        for key in copy_keys:
            copy_dict[key] = getattr(self, key)
        return copy_dict

    def __reset_local_vars(self, copy_dict):
        """ Resets the class variables in copy_dict. """
        for key, value in copy_dict.items():
            setattr(self, key, value)

    def __calc_param_indices(self, dims):
        indices_list = []
        for i in range(len(dims)):
            chunk = int(np.prod(dims[0:i]))
            repeat = int(np.prod(dims[i+1:]))
            idx = np.ravel(np.kron(list(range(dims[i])), np.ones((repeat, chunk))))
            indices_list.append(idx.astype(int))
        return np.transpose(np.array(indices_list))

    def __set_communicator(self, comm):
        self._communicator = comm

    def get_communicator(self):
        return self._communicator



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 .. module:: looks for a specific value in the provided second array (e.g. a mask) and substitutes the value in the first array with a given value
   :platform: Unix
   :synopsis: looks for a specific value in the provided second array (e.g. a mask) and substitutes the value in the first array with a given value

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""


from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np

@register_plugin
class ValueMaskReplacement(Plugin, CpuPlugin):
    """
    The function looks for a specific value in the provided second array (e.g. a mask) and substitutes the value in the first array with a given value. 

    :param seek_value: The value to be located in the provided second array. Default: 0.
    :param new_value: The value to be replaced with in the first array. Default: 1.   
    """

    def __init__(self):
        super(ValueMaskReplacement, self).__init__("ValueMaskReplacement")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_XZ', 'single')
        in_pData[1].plugin_data_setup('VOLUME_XZ', 'single')

        out_dataset[0].create_dataset(in_dataset[0])
        out_pData[0].plugin_data_setup('VOLUME_XZ', 'single')

    def process_frames(self, data):
        indeces = np.where(data[1] == self.parameters['seek_value'])
        corr_data = np.copy(data[0])
        corr_data[indeces] = self.parameters['new_value']
        return corr_data

    def nInput_datasets(self):
        return 2

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: docstring_parser
   :platform: Unix
   :synopsis: Parser for the docstring to interpret parameters and plugin info.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import sys
import re


def find_args(dclass, inst=None):
    """
    Finds the parameters list from the docstring
    """
    docstring = None
    if not dclass.__doc__:
        if inst:
            inst._override_class_docstring()
            docstring = dclass._override_class_docstring.__doc__
    else:
        docstring = dclass.__doc__

    if not docstring:
        return []

    mod_doc_lines = _get_doc_lines(sys.modules[dclass.__module__].__doc__)
    lines = _get_doc_lines(docstring)

    return _parse_args(mod_doc_lines, lines)


def _parse_args(mod_doc_lines, lines):
    param_list, user, hide, not_param, param_lines = __get_params(lines)

    warn_regexp = re.compile(r'^:config_warn: \s?(?P<config_warn>.*[^ ])$')
    warn, idx1 = __find_regexp(warn_regexp, lines)
    warn = '' if not warn else '.\n'.join(warn)+'.'
    syn_regexp = re.compile(r'^:synopsis: \s?(?P<synopsis>.*[^ ])$')
    synopsis, idx2 = __find_regexp(syn_regexp, mod_doc_lines)
    synopsis = '' if not synopsis else synopsis[0]+'.'

    info = __find_docstring_info(param_lines+idx1+idx2, lines)

    return {'warn': warn, 'info': info, 'synopsis': synopsis,
            'param': param_list, 'hide_param': hide, 'user_param': user,
            'not_param': not_param}


def _get_doc_lines(doc):
    if not doc:
        return ['']
    return [" ".join(l.strip(' .').split()) for l in doc.split('\n')]


def __get_params(lines):
    fixed_str = '(?P<param>\w+):\s?(?P<doc>\w.*[^ ])\s'\
                + '?Default:\s?(?P<default>.*[^ ])$'
    param_regexp = re.compile('^:param ' + fixed_str)
    param, idx1 = __find_regexp(param_regexp, lines)

    not_param_regexp = re.compile('^:~param (?P<param>\w+):')
    not_param, idx2 = __find_regexp(not_param_regexp, lines)

    hidden_param_regexp = re.compile('^:\*param ' + fixed_str)
    hidden_param, idx3 = __find_regexp(hidden_param_regexp, lines)
    hide_keys = [p[0] for p in hidden_param]

    user_param_regexp = re.compile('^:u\*param ' + fixed_str)
    user_param, idx4 = __find_regexp(user_param_regexp, lines)
    user_keys = [p[0] for p in user_param]

    lines_read = idx1+idx2+idx3+idx4

    param = user_param + param + hidden_param

    param_entry = [{'dtype': type(value), 'name': a[0], 'desc': a[1],
                    'default': value} for a in param for value in [eval(a[2])]]

    return param_entry, user_keys, hide_keys, not_param, lines_read


def __find_regexp(regexp, str_list):
    args = [regexp.findall(s) for s in str_list]
    index = [i for i in range(len(args)) if args[i]]
    args = [arg[0] for arg in args if len(arg)]
    return args, index


def __find_docstring_info(index, str_list):
    info = [str_list[i] for i in range(len(str_list)) if i not in index]
    info = [i for i in info if i]
    info = "" if not info else "\n".join(info) + '.'
    return info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: content
   :platform: Unix
   :synopsis: Content class for the configurator

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import re
import os
import inspect

from savu.plugins import utils as pu
from savu.data.plugin_list import PluginList
from . import mutations


class Content(object):

    def __init__(self, filename=None, level='user'):
        self.disp_level = level
        self.plugin_list = PluginList()
        self.plugin_mutations = mutations.plugin_mutations
        self.param_mutations = mutations.param_mutations
        self.filename = filename
        self._finished = False
        self.failed = {}

    def set_finished(self, check='y'):
        self._finished = True if check.lower() == 'y' else False

    def is_finished(self):
        return self._finished

    def fopen(self, infile, update=False, skip=False):
        if os.path.exists(infile):
            self.plugin_list._populate_plugin_list(infile, active_pass=True)
        else:
            raise Exception('INPUT ERROR: The file does not exist.')
        self.filename = infile
        if update:
            self.plugin_mutations = self.check_mutations(self.plugin_mutations)
            self.param_mutations = self.check_mutations(self.param_mutations)
            self._apply_plugin_updates(skip)

    def check_mutations(self, mut_dict: dict):
        plist_version = self._version_to_float(self.plugin_list.version)
        # deleting elements while iterating invalidates the iterator
        # which raises a RuntimeError in Python 3.
        # Instead a copy of the dict is mutated and returned
        mut_dict_copy = mut_dict.copy()
        for key, subdict in mut_dict.items():
            if 'up_to_version' in subdict.keys():
                up_to_version = self._version_to_float(subdict['up_to_version'])
                if plist_version >= up_to_version:
                    del mut_dict_copy[key]
        return mut_dict_copy

    def _version_to_float(self, version):
        if version is None:
            return 0
        if isinstance(version, bytes):
            version = version.decode("ascii")
        split_vals = version.split('.')
        return float('.'.join([split_vals[0], ''.join(split_vals[1:])]))

    def display(self, formatter, **kwargs):
        if 'level' not in list(kwargs.keys()):
            kwargs['level'] = self.disp_level
        print('\n' + formatter._get_string(**kwargs) + '\n')

    def check_file(self, filename):
        if not filename:
            raise Exception('INPUT ERROR: Please specify the output filepath.')
        path = os.path.dirname(filename)
        path = path if path else '.'
        if not os.path.exists(path):
            file_error = "INPUT_ERROR: Incorrect filepath."
            raise Exception(file_error)

    def save(self, filename, check='y', template=False):
        if check.lower() == 'y':
            print(f"Saving file {filename}")
            if template:
                self.plugin_list.add_template(create=True)
            self.plugin_list._save_plugin_list(filename)
        else:
            print("The process list has NOT been saved.")

    def clear(self, check='y'):
        if check.lower() == 'y':
            self.plugin_list.plugin_list = []

    def add(self, name, str_pos):
        if name not in list(pu.plugins.keys()):
            if name in list(self.failed.keys()):
                msg = "IMPORT ERROR: %s is unavailable due to the following" \
                      " error:\n\t%s" % (name, self.failed[name])
                raise Exception(msg)
            else:
                raise Exception("INPUT ERROR: Unknown plugin %s" % name)
        plugin = pu.plugins[name]()
        plugin._populate_default_parameters()
        pos, str_pos = self.convert_pos(str_pos)
        self.insert(plugin, pos, str_pos)

    def refresh(self, str_pos, defaults=False, change=False):
        pos = self.find_position(str_pos)
        plugin_entry = self.plugin_list.plugin_list[pos]
        name = change if change else plugin_entry['name']
        active = plugin_entry['active']
        plugin = pu.plugins[name]()
        plugin._populate_default_parameters()

        keep = self.get(pos)['data'] if not defaults else None
        self.insert(plugin, pos, str_pos, replace=True)
        self.plugin_list.plugin_list[pos]['active'] = active
        if keep:
            self._update_parameters(plugin, name, keep, str_pos)

    def _update_parameters(self, plugin, name, keep, str_pos):
        union_params = set(keep).intersection(set(plugin.parameters))
        for param in union_params:
            self.modify(str_pos, param, keep[param], ref=True)
        # add any parameter mutations here
        classes = [c.__name__ for c in inspect.getmro(plugin.__class__)]
        m_dict = self.param_mutations
        keys = [k for k in list(m_dict.keys()) if k in classes]

        changes = False
        for k in keys:
            for entry in m_dict[k]['params']:
                if entry['old'] in list(keep.keys()):
                    changes = True
                    val = keep[entry['old']]
                    if 'eval' in list(entry.keys()):
                        val = eval(entry['eval'])
                    self.modify(str_pos, entry['new'], val, ref=True)
        if changes:
            mutations.param_change_str(keep, plugin.parameters, name, keys)

    def _apply_plugin_updates(self, skip=False):
        # Update old process lists that start from 0
        the_list = self.plugin_list.plugin_list
        if 'pos' in list(the_list[0].keys()) and the_list[0]['pos'] == '0':
            self.increment_positions()

        missing = []
        pos = len(the_list) - 1
        notices = mutations.plugin_notices

        for plugin in the_list[::-1]:
            # update old process lists to include 'active' flag
            if 'active' not in list(plugin.keys()):
                plugin['active'] = True

            while True:
                name = the_list[pos]['name']
                if name in notices.keys():
                    print(notices[name]['desc'])

                # if a plugin is missing from all available plugins
                # then look for mutations in the plugin name
                search = True if name not in pu.plugins.keys() else False
                found = self._mutate_plugins(name, pos, search=search)
                if search and not found:
                    str_pos = self.plugin_list.plugin_list[pos]['pos']
                    missing.append([name, str_pos])
                    self.remove(pos)
                    pos -= 1
                if name == the_list[pos]['name']:
                    break
            pos -= 1

        for name, pos in missing[::-1]:
            if skip:
                print(f"Skipping plugin {pos}: {name}")
            else:
                message = f"PLUGIN ERROR: The plugin {name} is unavailable in this version of Savu."
                print(message)
                raise Exception(f'Incompatible process list. {message}')

    def _mutate_plugins(self, name, pos, search=False):
        """ Perform plugin mutations. """
        # check for case changes in plugin name
        if search:
            for key in pu.plugins.keys():
                if name.lower() == key.lower():
                    str_pos = self.plugin_list.plugin_list[pos]['pos']
                    self.refresh(str_pos, change=key)
                    return True

        # check mutations dict
        m_dict = self.plugin_mutations
        if name in m_dict.keys():
            mutate = m_dict[name]
            if 'replace' in mutate.keys():
                if mutate['replace'] in list(pu.plugins.keys()):
                    str_pos = self.plugin_list.plugin_list[pos]['pos']
                    self.refresh(str_pos, change=mutate['replace'])
                    print(mutate['desc'])
                    return True
                raise Exception(f"Replacement plugin {mutate['replace']} unavailable for {name}")
            elif 'remove' in mutate.keys():
                self.remove(pos)
                print(mutate['desc'])
            else:
                raise Exception('Unknown mutation type.')
        return False

    def move(self, old, new):
        old_pos = self.find_position(old)
        entry = self.plugin_list.plugin_list[old_pos]
        self.remove(old_pos)
        new_pos, new = self.convert_pos(new)
        name = entry['name']
        self.insert(pu.plugins[name](), new_pos, new)
        self.plugin_list.plugin_list[new_pos] = entry
        self.plugin_list.plugin_list[new_pos]['pos'] = new

    def modify(self, pos_str, subelem, value, ref=False):
        if not ref:
            value = self.value(value)
        pos = self.find_position(pos_str)
        data_elements = self.plugin_list.plugin_list[pos]['data']
        if subelem.isdigit():
            subelem = self.plugin_list.plugin_list[pos]['map'][int(subelem) - 1]
        data_elements[subelem] = value

    def value(self, value):
        if not value.count(';'):
            try:
                value = eval(value)
            except (NameError, SyntaxError):
                value = eval(f"'{value}'")
        return value

    def convert_to_ascii(self, value):
        ascii_list = []
        for v in value:
            ascii_list.append(v.encode('ascii', 'ignore'))
        return ascii_list

    def on_and_off(self, str_pos, index):
        print(("switching plugin %s %s" % (str_pos, index)))
        status = True if index == 'ON' else False
        pos = self.find_position(str_pos)
        self.plugin_list.plugin_list[pos]['active'] = status

    def convert_pos(self, str_pos):
        """ Converts the display position (input) to the equivalent numerical
        position and updates the display position if required.

        :param str_pos: the plugin display position (input) string.
        :returns: the equivalent numerical position of str_pos and and updated\
            str_pos.
        :rtype: (pos, str_pos)
        """
        pos_list = self.get_split_positions()
        num = re.findall(r"\d+", str_pos)[0]
        letter = re.findall("[a-z]", str_pos)
        entry = [num, letter[0]] if letter else [num]

        # full value already exists in the list
        if entry in pos_list:
            index = pos_list.index(entry)
            return self.inc_positions(index, pos_list, entry, 1)

        # only the number exists in the list
        num_list = [pos_list[i][0] for i in range(len(pos_list))]
        if entry[0] in num_list:
            start = num_list.index(entry[0])
            if len(entry) is 2:
                if len(pos_list[start]) is 2:
                    idx = int([i for i in range(len(num_list)) if
                               (num_list[i] == entry[0])][-1]) + 1
                    entry = [entry[0], str(chr(ord(pos_list[idx - 1][1]) + 1))]
                    return idx, ''.join(entry)
                if entry[1] == 'a':
                    self.plugin_list.plugin_list[start]['pos'] = entry[0] + 'b'
                    return start, ''.join(entry)
                else:
                    self.plugin_list.plugin_list[start]['pos'] = entry[0] + 'a'
                    return start + 1, entry[0] + 'b'
            return self.inc_positions(start, pos_list, entry, 1)

        # number not in list
        entry[0] = str(int(num_list[-1]) + 1 if num_list else 1)
        if len(entry) is 2:
            entry[1] = 'a'
        return len(self.plugin_list.plugin_list), ''.join(entry)

    def increment_positions(self):
        """ Update old process lists that start plugin numbering from 0 to
        start from 1. """
        for plugin in self.plugin_list.plugin_list:
            str_pos = plugin['pos']
            num = str(int(re.findall(r'\d+', str_pos)[0]) + 1)
            letter = re.findall("[a-z]", str_pos)
            plugin['pos'] = ''.join([num, letter[0]] if letter else [num])

    def get_positions(self):
        """ Get a list of all current plugin entry positions. """
        elems = self.plugin_list.plugin_list
        pos_list = []
        for e in elems:
            pos_list.append(e['pos'])
        return pos_list

    def get_split_positions(self):
        """ Separate numbers and letters in positions. """
        positions = self.get_positions()
        split_pos = []
        for i in range(len(positions)):
            num = re.findall(r'\d+', positions[i])[0]
            letter = re.findall('[a-z]', positions[i])
            split_pos.append([num, letter[0]] if letter else [num])
        return split_pos

    def find_position(self, pos):
        """ Find the numerical index of a position (a string). """
        pos_list = self.get_positions()
        if pos not in pos_list:
            raise Exception("INPUT ERROR: Incorrect plugin position.")
        return pos_list.index(pos)

    def inc_positions(self, start, pos_list, entry, inc):
        if len(entry) is 1:
            self.inc_numbers(start, pos_list, inc)
        else:
            idx = [i for i in range(start, len(pos_list)) if
                   pos_list[i][0] == entry[0]]
            self.inc_letters(idx, pos_list, inc)
        return start, ''.join(entry)

    def inc_numbers(self, start, pos_list, inc):
        for i in range(start, len(pos_list)):
            pos_list[i][0] = str(int(pos_list[i][0]) + inc)
            self.plugin_list.plugin_list[i]['pos'] = ''.join(pos_list[i])

    def inc_letters(self, idx, pos_list, inc):
        for i in idx:
            pos_list[i][1] = str(chr(ord(pos_list[i][1]) + inc))
            self.plugin_list.plugin_list[i]['pos'] = ''.join(pos_list[i])

    def insert(self, plugin, pos, str_pos, replace=False):
        plugin_dict = self.create_plugin_dict(plugin)
        plugin_dict['pos'] = str_pos
        if replace:
            self.plugin_list.plugin_list[pos] = plugin_dict
        else:
            self.plugin_list.plugin_list.insert(pos, plugin_dict)

    def create_plugin_dict(self, plugin):
        plugin_dict = {}
        plugin_dict['name'] = plugin.name
        plugin_dict['id'] = plugin.__module__
        plugin_dict['data'] = plugin.parameters
        plugin_dict['active'] = True
        plugin_dict['desc'] = plugin.parameters_desc
        plugin_dict['hide'] = plugin.parameters_hide
        plugin_dict['user'] = plugin.parameters_user

        dev_keys = [k for k in list(plugin_dict['data'].keys()) if k not in
                    plugin_dict['user'] + plugin_dict['hide']]
        plugin_dict['map'] = \
            plugin_dict['user'] + dev_keys + plugin_dict['hide']
        return plugin_dict

    def get(self, pos):
        return self.plugin_list.plugin_list[pos]

    def remove(self, pos):
        if pos >= self.size:
            raise Exception("Cannot remove plugin %s as it does not exist."
                            % self.plugin_list.plugin_list[pos]['name'])
        pos_str = self.plugin_list.plugin_list[pos]['pos']
        self.plugin_list.plugin_list.pop(pos)
        pos_list = self.get_split_positions()
        self.inc_positions(pos, pos_list, pos_str, -1)

    @property
    def size(self):
        return len(self.plugin_list.plugin_list)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_component_analysis
   :platform: Unix
   :synopsis: A base class for all component analysis methods

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
import sys
import numpy as np


class BaseComponentAnalysis(Plugin, CpuPlugin):
    """
    A base plugin for doing component analysis. This sorts out the main \
    features of a component analysis

    :param in_datasets: A list of the dataset(s) to process. Default: [].
    :param out_datasets: A list of the dataset(s) to \
        process. Default: ['scores', 'eigenvectors'].
    :param number_of_components: The number expected components. Default: 3.
    :param chunk: The chunk to work on. Default: 'SINOGRAM'.
    :param whiten: To subtract the mean or not. Default: 1.
    """

    def __init__(self, name):
        super(BaseComponentAnalysis, self).__init__(name)

    def get_max_frames(self):
        return self.spectra_length[0]

    def get_plugin_pattern(self):
        return self.parameters['chunk']

    def setup(self):
        self.exp.log(self.name + " Setting up the component analysis")
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()
        self.spectra_length = (in_dataset[0].get_shape()[-1],)
        other_dims = in_dataset[0].get_shape()[:-1]
        num_comps = self.parameters['number_of_components']
        self.images_shape = other_dims + (num_comps,)
        components_shape = (num_comps,) + self.spectra_length
        # copy all required information from in_dataset[0]

        out_dataset[0].create_dataset(patterns=in_dataset[0],
                                      axis_labels=in_dataset[0],
                                      shape=self.images_shape)

        axis_labels = ['idx.unit', 'spectra.unit']
        out_dataset[1].create_dataset(shape=components_shape,
                                      axis_labels=axis_labels)
        spectrum = {'core_dims': (1,), 'slice_dims': (0,)}

        out_dataset[1].add_pattern("SPECTRUM", **spectrum)

        in_pData, out_pData = self.get_plugin_datasets()
        plugin_pattern = self.get_plugin_pattern()
#         dirs = range(len(out_dataset[0].get_shape()))
#         vxz = {'core_dims': (0,1), 'slice_dims': (2,)}
#         in_dataset[0].add_pattern("VOLUME_XZ", **vxz)
        in_pData[0].plugin_data_setup(plugin_pattern, self.get_max_frames())
        out_pData[0].plugin_data_setup(plugin_pattern, num_comps)
        out_pData[1].plugin_data_setup("SPECTRUM", num_comps)

        self.exp.log(self.name + " End")

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 2

    def remove_nan_inf(self, data):
        '''
        converts the nans to nums and sets the infs to the max float size
        not strictly true, but does allow fitting to take place
        '''
        data = np.clip(data,-sys.float_info.max,sys.float_info.max)
        data[np.isnan(data)]=0
        data = np.nan_to_num(data)
        return data



# Copyright 2018 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomobar_recon
   :platform: Unix
   :synopsis: A wrapper around TOmographic MOdel-BAsed Reconstruction (ToMoBAR) software \
   for advanced iterative image reconstruction (2D case) using GPU

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.gpu_plugin import GpuPlugin

import numpy as np
from tomobar.methodsIR import RecToolsIR

from savu.plugins.utils import register_plugin

@register_plugin
class TomobarRecon(BaseRecon, GpuPlugin):
    """
    A GPU plugin to reconstruct full-field tomographic projection data using state-of-the-art regularised iterative algorithms from \
    the ToMoBAR package. ToMoBAR includes FISTA and ADMM iterative methods and depends on the ASTRA toolbox and the CCPi RGL toolkit: \
    https://github.com/vais-ral/CCPi-Regularisation-Toolkit.

    :param output_size: Number of rows and columns in the \
        reconstruction. Default: 'auto'.
    :param data_fidelity: Data fidelity, choose LS, PWLS, SWLS or KL. Default: 'LS'.
    :param data_Huber_thresh: Threshold parameter for __Huber__ data fidelity . Default: None.
    :param data_beta_SWLS: A parameter for stripe-weighted model. Default: 0.1.
    :param data_full_ring_GH: Regularisation variable for full constant ring removal (GH model). Default: None.
    :param data_full_ring_accelerator_GH: Acceleration constant for GH ring removal. Default: 10.0.
    :param algorithm_iterations: Number of outer iterations for FISTA (default) or ADMM methods. Default: 20.
    :param algorithm_verbose: print iterations number and other messages ('off' by default). Default: 'off'.
    :param algorithm_mask: set to 1.0 to enable a circular mask diameter or < 1.0 to shrink the mask. Default: 1.0.
    :param algorithm_ordersubsets: The number of ordered-subsets to accelerate reconstruction. Default: 6.
    :param algorithm_nonnegativity: ENABLE or DISABLE nonnegativity constraint. Default: 'ENABLE'.
    :param regularisation_method: To regularise choose methods ROF_TV, FGP_TV, PD_TV, SB_TV, LLT_ROF,\
                             NDF, TGV, NLTV, Diff4th. Default: 'FGP_TV'.
    :param regularisation_parameter: Regularisation (smoothing) value, higher \
                            the value stronger the smoothing effect. Default: 0.00001.
    :param regularisation_iterations: The number of regularisation iterations. Default: 80.
    :param regularisation_PD_lip: Primal-dual parameter for convergence. Default: 8.
    :param regularisation_methodTV:  0/1 - TV specific isotropic/anisotropic choice. Default: 0.
    :param regularisation_timestep: Time marching parameter, relevant for \
                    (ROF_TV, LLT_ROF, NDF, Diff4th) penalties. Default: 0.003.
    :param regularisation_edge_thresh: Edge (noise) related parameter, relevant for NDF and Diff4th. Default: 0.01.
    :param regularisation_parameter2:  Regularisation (smoothing) value for LLT_ROF method. Default: 0.005.
    :param regularisation_NDF_penalty: NDF specific penalty type Huber, Perona, Tukey. Default: 'Huber'.
    """

    def __init__(self):
        super(TomobarRecon, self).__init__("TomobarRecon")

    def pre_process(self):
        # extract given parameters into dictionaries suitable for ToMoBAR input
        """
        # current parameters not fully working yet
        :param data_any_rings: a parameter to suppress various artifacts including rings and streaks. Default: None.
        :param data_any_rings_winsizes: half window sizes to collect background information [detector, angles, num of projections]. Default: (9,7,0).
        :param data_any_rings_power: a power parameter for Huber model. Default: 1.5.
        'ring_weights_threshold' :  self.parameters['data_any_rings'],
        'ring_tuple_halfsizes' :  self.parameters['data_any_rings_winsizes'],
        'ring_huber_power' :  self.parameters['data_any_rings_power'],
        """
        self._data_ = {'OS_number' : self.parameters['algorithm_ordersubsets'],
                       'huber_threshold' : self.parameters['data_Huber_thresh'],
                       'ringGH_lambda' :  self.parameters['data_full_ring_GH'],
                       'ringGH_accelerate' :  self.parameters['data_full_ring_accelerator_GH']}

        self._algorithm_ = {'iterations' : self.parameters['algorithm_iterations'],
			                'nonnegativity' : self.parameters['algorithm_nonnegativity'],
                            'mask_diameter' : self.parameters['algorithm_mask'],
                            'verbose' : self.parameters['algorithm_verbose']}

        self._regularisation_ = {'method' : self.parameters['regularisation_method'],
                                'regul_param' : self.parameters['regularisation_parameter'],
                                'iterations' : self.parameters['regularisation_iterations'],
                                'device_regulariser' : 'gpu',
                                'edge_threhsold' : self.parameters['regularisation_edge_thresh'],
                                'time_marching_step' : self.parameters['regularisation_timestep'],
                                'regul_param2' : self.parameters['regularisation_parameter2'],
                                'PD_LipschitzConstant' : self.parameters['regularisation_PD_lip'],
                                'NDF_penalty' : self.parameters['regularisation_NDF_penalty'],
                                'methodTV' : self.parameters['regularisation_methodTV']}

    def process_frames(self, data):
        cor, angles, self.vol_shape, init  = self.get_frame_params()
        sinogram = data[0].astype(np.float32)
        anglesTot, self.DetectorsDimH = np.shape(sinogram)
        half_det_width = 0.5*self.DetectorsDimH
        cor_astra = half_det_width - cor
        self.anglesRAD = np.deg2rad(angles.astype(np.float32))
        self._data_.update({'projection_norm_data' : sinogram})

        # if one selects PWLS or SWLS models then raw data is also required (2 inputs)
        if ((self.parameters['data_fidelity'] == 'PWLS') or (self.parameters['data_fidelity'] == 'SWLS')):
            rawdata = data[1].astype(np.float32)
            rawdata /= np.max(rawdata)
            self._data_.update({'projection_raw_data' : rawdata})
            self._data_.update({'beta_SWLS' : self.parameters['data_beta_SWLS']*np.ones(self.DetectorsDimH)})

        # set parameters and initiate the ToMoBAR class object
        self.Rectools = RecToolsIR(DetectorsDimH = self.DetectorsDimH,  # DetectorsDimH # detector dimension (horizontal)
                    DetectorsDimV = None,  # DetectorsDimV # detector dimension (vertical) for 3D case only
                    CenterRotOffset = cor_astra.item() - 0.5, # The center of rotation (CoR) scalar or a vector
                    AnglesVec = self.anglesRAD, # the vector of angles in radians
                    ObjSize = self.vol_shape[0] , # a scalar to define the reconstructed object dimensions
                    datafidelity=self.parameters['data_fidelity'],# data fidelity, choose LS, PWLS, SWLS
                    device_projector='gpu')

        # Run FISTA reconstrucion algorithm here
        recon = self.Rectools.FISTA(self._data_, self._algorithm_, self._regularisation_)
        return recon

    def nInput_datasets(self):
        return max(len(self.parameters['in_datasets']), 1)

    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info1 = CitationInformation()
        cite_info1.name = 'citation1'
        cite_info1.description = \
            ("First-order optimisation algorithm for linear inverse problems.")
        cite_info1.bibtex = \
            ("@article{beck2009,\n" +
             "title={A fast iterative shrinkage-thresholding algorithm for linear inverse problems},\n" +
             "author={Amir and Beck, Mark and Teboulle},\n" +
             "journal={SIAM Journal on Imaging Sciences},\n" +
             "volume={2},\n" +
             "number={1},\n" +
             "pages={183--202},\n" +
             "year={2009},\n" +
             "publisher={SIAM}\n" +
             "}")
        cite_info1.endnote = \
            ("%0 Journal Article\n" +
             "%T A fast iterative shrinkage-thresholding algorithm for linear inverse problems\n" +
             "%A Beck, Amir\n" +
             "%A Teboulle, Mark\n" +
             "%J SIAM Journal on Imaging Sciences\n" +
             "%V 2\n" +
             "%N 1\n" +
             "%P 183--202\n" +
             "%@ --\n" +
             "%D 2009\n" +
             "%I SIAM\n")
        cite_info1.doi = "doi: "
        return cite_info1



# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: test_utils
   :platform: Unix
   :synopsis: utilities for the test framework

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import inspect
import tempfile
import os
import copy
import glob
import shutil

from savu.core.plugin_runner import PluginRunner
from savu.data.experiment_collection import Experiment
from savu.data.data_structures.plugin_data import PluginData
import savu.plugins.utils as pu
import savu.test.base_checkpoint_test


def get_test_data_path(name):
    path = inspect.stack()[0][1]
    return '/'.join(os.path.split(path)[0].split(os.sep)[:-2] +
                    ['test_data/data', name])


def get_test_big_data_path(name):
    '''
    internal to diamond
    '''
    path = '/dls/mx-scratch/savu_test_data'
    return path+os.sep+name


def get_test_process_path(name):
    path = inspect.stack()[0][1]
    full_path = '/'.join(os.path.split(path)[0].split(os.sep)[:-2] +
                         ['test_data/test_process_lists', name])
    return full_path


def get_process_list_path(name):
    path = inspect.stack()[0][1]
    return '/'.join(os.path.split(path)[0].split(os.sep)[:-2] +
                    ['test_data/process_lists', name])


def get_experiment_types():
    exp_dict = {}
    exp_dict['tomoRaw'] = {'func': 'set_tomoRaw_experiment',
                           'filename': '24737.nxs'}
    exp_dict['tomo'] = {'func': 'set_tomo_experiment',
                        'filename': '24737_processed.nxs'}
    exp_dict['fluo'] = {'func': 'set_fluo_experiment',
                        'filename': 'fluo.nxs'}
    exp_dict['tomo_3dto4d'] = {'func': 'set_3dto4d_experiment',
                               'filename': 'i12_test_data.nxs'}

    return exp_dict


def set_experiment(exp_type, **kwargs):
    exp_types = get_experiment_types()
    try:
        options = globals()[exp_types[exp_type]['func']](
            exp_types[exp_type]['filename'], **kwargs)
    except KeyError:
        raise Exception("The experiment type ", exp_type, " is not recognised")
    return options


def set_tomoRaw_experiment(filename, **kwargs):
    # create experiment
    options = set_options(get_test_data_path(filename))
    options['loader'] = 'savu.plugins.loaders.full_field_loaders.nxtomo_loader'
    return options


def set_tomo_experiment(filename, **kwargs):
    options = set_options(get_test_data_path('/full_field_corrected/' + filename), **kwargs)
    options['loader'] = 'savu.plugins.loaders.savu_nexus_loader'
    return options


def set_fluo_experiment(filename, **kwargs):
    options = set_options(get_test_data_path(filename), **kwargs)
    options['loader'] = 'savu.plugins.loaders.mapping_loaders.nxfluo_loader'
    return options


def set_3dto4d_experiment(filename, **kwargs):
    options = set_options(
        get_test_data_path('/i12_test_data/' + filename), **kwargs)
    options['loader'] = 'savu.plugins.loaders.full_field_loaders.nxtomo_loader'
    return options


def get_output_datasets(plugin):
    n_out = plugin.nOutput_datasets()
    out_data = []
    if n_out is 'var':
        return None
    for n in range(n_out):
        out_data.append('test' + str(n))
    return out_data


def set_plugin_list(options, pnames, *args):
    args = args[0] if args else None
    plugin_names = pnames if isinstance(pnames, list) else [pnames]
    options['plugin_list'] = []
    ID = [options['loader']]
    data = [{}, {}] if not args else [args[0], args[-1]]
    for i in range(len(plugin_names)):
        ID.insert(i+1, plugin_names[i])
        plugin = pu.load_class(plugin_names[i])()
        data_dict = set_data_dict(['tomo'], get_output_datasets(plugin))
        data_dict = args[i+1] if args else data_dict
        data.insert(i+1, data_dict)

    for i in range(len(ID)):
        name = \
            ''.join(x.capitalize() for x in (ID[i].split('.')[-1]).split('_'))
        options['plugin_list'].append(set_plugin_entry(
            name, ID[i], data[i], i))


def set_plugin_entry(name, ID, data, pos):
    plugin = {}
    plugin['name'] = name
    plugin['id'] = ID
    plugin['data'] = data
    plugin['desc'] = data
    plugin['hide'] = []
    plugin['user'] = []
    plugin['active'] = []
    plugin['pos'] = str(pos)
    return plugin


def set_options(path, **kwargs):
    options = {}
    options['transport'] = kwargs.get('transport', 'hdf5')
    options['process_names'] = kwargs.get('process_names', 'CPU0')
    options['nProcesses'] = len(options["process_names"].split(','))
    options['data_file'] = path
    options['process_file'] = kwargs.get('process_file', '')
    options['out_path'] = kwargs.get('out_path', tempfile.mkdtemp())
    options['out_folder'] = 'test'
    options['datafile_name'] = 'test'
    options['inter_path'] = options['out_path']
    options['log_path'] = options['out_path']
    options['run_type'] = 'test'
    options['verbose'] = 'True'
    #options['link_type'] = 'final_result'
    options['test_state'] = True
    options['lustre'] = False
    options['bllog'] = None
    options['email'] = None
    options['template'] = None
    options['checkpoint'] = None
    options['system_params'] = None
    options['nPlugin'] = 0
    return options


def set_data_dict(in_data, out_data):
    return {'in_datasets': in_data, 'out_datasets': out_data}


def _add_loader_to_plugin_list(options, params={}):
    plugin_list = []
    ID = options['loader']
    name = ''.join(x.capitalize() for x in (ID.split('.')[-1]).split('_'))
    plugin_list.append(set_plugin_entry(name, ID, params, 0))
    options['plugin_list'] = plugin_list


def load_random_data(loader, params, system_params=None, fake=False):
    options = set_options(get_test_data_path('24737.nxs'))
    options['loader'] = 'savu.plugins.loaders.' + str(loader)
    options['system_params'] = system_params
    _add_loader_to_plugin_list(options, params=params)
    if fake:
        return fake_plugin_runner(options).exp
    return plugin_runner(options)


def load_test_data(exp_type):
    options = set_experiment(exp_type)
    _add_loader_to_plugin_list(options)
    return plugin_runner(options)


def get_data_object(exp):
    data = exp.index['in_data'][list(exp.index['in_data'].keys())[0]]
    data._set_plugin_data(PluginData(data))
    pData = data._get_plugin_data()
    return data, pData


def set_process(exp, process, processes):
    exp.meta_data.set('process', process)
    exp.meta_data.set('processes', processes)


def plugin_runner(options):
    # the real plugin runner runs a full plugin list
    plugin_runner = PluginRunner(options)
    return plugin_runner._run_plugin_list()


def fake_plugin_runner(options):
    # Stripped down version of the plugin runner
    # Loads the loader plugin but stops before processing plugins
    plugin_runner = PluginRunner(options)
    plugin_runner.exp = Experiment(options)
    plugin_list = plugin_runner.exp.meta_data.plugin_list.plugin_list

    exp = plugin_runner.exp
    pu.plugin_loader(exp, plugin_list[0])
    exp._set_nxs_file()
    return plugin_runner


def plugin_runner_load_plugin(options):
    # Loads the loader plugin and the first plugin in the list and initialises
    # it (i.e. sets parameters and calls setup method)
    pRunner = fake_plugin_runner(options)
    plugin_list = pRunner.exp.meta_data.plugin_list.plugin_list
    plugin_dict = plugin_list[1]
    plugin = pu.plugin_loader(pRunner.exp, plugin_dict)
    return plugin


def plugin_runner_real_plugin_run(options):
    plugin_runner = PluginRunner(options)
    plugin_list_inst = plugin_runner.exp.meta_data.plugin_list
    plugin_list = plugin_list_inst.plugin_list
    plugin_runner._run_plugin_list_check(plugin_list_inst)

    exp = plugin_runner.exp
    pu.plugin_loader(exp, plugin_list[0])

    start_in_data = copy.deepcopy(exp.index['in_data'])
    in_data = exp.index["in_data"][list(exp.index["in_data"].keys())[0]]
    out_data_objs, stop = in_data._load_data(1)
    exp._clear_data_objects()
    exp.index['in_data'] = copy.deepcopy(start_in_data)

    for key in out_data_objs[0]:
        exp.index["out_data"][key] = out_data_objs[0][key]

    plugin = pu.plugin_loader(exp, plugin_list[1])
    plugin._run_plugin(exp, plugin_runner)


def initialise_options(data, experiment, process_path):
    """
    initialises options and creates a temporal directory in tmp for output.

    :param str data: data to run test with (can be None)
    :param str experiment: experiment type to run test with (can be None)
    :param str process_path: a path to the preocess list (can be None)
    """
    test_folder = tempfile.mkdtemp(suffix='my_test/')
    if data is not None:
        data_file = get_test_data_path(data)
    if process_path is not None:
        process_file = get_test_process_path(process_path)
    if (experiment is not None) & (data is None):
        options = set_experiment(experiment)
    elif (experiment is not None) & (data is not None):
        if experiment == 'load_data':
            options = set_experiment('tomo')
            options['data_file'] = data
        else:
            options = set_experiment(experiment)
            options['data_file'] = data_file
        options['process_file'] = process_file
    else:
        options = set_options(data_file, process_file=process_file)
    options['out_path'] = os.path.join(test_folder)
    return options


def cleanup(options):
    """
    Performs folders cleaning in tmp/.
    using _shutil_ module in order to delete everything recursively
    """
    shutil.rmtree(options["out_path"], ignore_errors=True)
    """
    classb = savu.test.base_checkpoint_test.BaseCheckpointTest()
    cp_folder = os.path.join(options["out_path"], 'checkpoint')
    classb._empty_folder(cp_folder)
    # delete folders after imagesavers
    im_folder = os.path.join(options["out_path"], 'ImageSaver-tomo')
    if os.path.isdir(im_folder):
        classb._empty_folder(im_folder)
        os.removedirs(im_folder)
    im_folder = os.path.join(options["out_path"], 'TiffSaver-tomo')
    if os.path.isdir(im_folder):
        classb._empty_folder(im_folder)
        os.removedirs(im_folder)
    os.removedirs(cp_folder)
    classb._empty_folder(options["out_path"])
    os.removedirs(options["out_path"])
    """
    return options



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: pyfai_azimuthal_integrator
   :platform: Unix
   :synopsis: A plugin to integrate azimuthally "symmetric" signals i.e. SAXS,\
       WAXS or XRD.Requires a calibration file


.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
from savu.plugins.azimuthal_integrators.base_azimuthal_integrator import \
    BaseAzimuthalIntegrator

from savu.plugins.utils import register_plugin


@register_plugin
class PyfaiAzimuthalIntegrator(BaseAzimuthalIntegrator):
    """
    1D azimuthal integrator by pyFAI
    :param use_mask: Should we mask. Default: False.
    :param num_bins: number of bins. Default: 1005.
    """

    def __init__(self):
        logging.debug("Starting 1D azimuthal integrationr")
        super(PyfaiAzimuthalIntegrator,
              self).__init__("PyfaiAzimuthalIntegrator")

    def process_frames(self, data):
        logging.debug("Running azimuthal integration")
        ai = self.params[3]
        logging.debug('datashape=%s' % str(data[0].shape))
        axis, remapped = \
            ai.integrate1d(data=data[0], npt=self.npts, unit='q_A^-1',
                           correctSolidAngle=False)
        return remapped



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_driver
   :platform: Unix
   :synopsis: Base class or all driver plugins

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import logging
from mpi4py import MPI


class BaseDriver(object):
    """
    The base class from which all driver plugins should inherit.
    """

    def get_communicator(self):
        """
        This method should return an MPI communicator
        """
        logging.error("get_communicator needs to be implemented")
        raise NotImplementedError("get_communicator needs to be implemented.")

    def plugin_barrier(self):
        return self.exp._barrier(communicator=self.get_communicator())

    def _run_plugin_instances(self, transport, communicator=MPI.COMM_WORLD):
        """ Runs the pre_process, process and post_process methods.

        If parameter tuning is required, loop over the methods and set the
        correct parameters for each run. """

        self.__set_communicator(communicator)

        logging.info("%s.%s", self.__class__.__name__, 'pre_process')
        self.base_pre_process()
        self.pre_process()
        logging.info("%s.%s", self.__class__.__name__, '_barrier')
        self.plugin_barrier()

        logging.info("%s.%s", self.__class__.__name__, 'transport_process_frames')
        transport._transport_checkpoint(transport._transport_process(self))

        logging.info("%s.%s", self.__class__.__name__, '_barrier')
        self.plugin_barrier()

        logging.info("%s.%s", self.__class__.__name__, 'post_process')
        self.post_process()
        self.base_post_process()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxfluo_loader
   :platform: Unix
   :synopsis: A class for loading nxfluo data

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.plugins.loaders.mapping_loaders.base_multi_modal_loader \
    import BaseMultiModalLoader

from savu.plugins.utils import register_plugin


@register_plugin
class NxfluoLoader(BaseMultiModalLoader):
    """ A class to load tomography data from an NXFluo file.

    :param fluo_offset: fluo scale offset. Default: 0.0.
    :param fluo_gain: fluo gain. Default: 0.01.
    :param name: The name assigned to the dataset. Default: 'fluo'.
    """

    def __init__(self, name='NxfluoLoader'):
        super(NxfluoLoader, self).__init__(name)

    def setup(self):
        path = 'instrument/fluorescence/data'
        data_obj, fluo_entry = \
            self.multi_modal_setup('NXfluo', path, self.parameters['name'])

        npts = data_obj.data.shape[self.get_motor_dims('None')[0]]
        gain = self.parameters["fluo_gain"]
        energy = np.arange(self.parameters["fluo_offset"], gain*npts, gain)
        mono_path = fluo_entry.name + '/instrument/monochromator/energy'
        mono_energy = data_obj.backing_file[mono_path][()]
        monitor_path = fluo_entry.name + '/monitor/data'
        monitor = data_obj.backing_file[monitor_path][()]

        data_obj.meta_data.set("energy", energy)
        data_obj.meta_data.set("mono_energy", mono_energy)
        data_obj.meta_data.set("monitor", monitor)
        data_obj.meta_data.set("average", np.ones(npts))

        slice_dir = tuple(range(len(data_obj.data.shape)-1))
        data_obj.add_pattern("SPECTRUM", core_dims=(-1,), slice_dims=slice_dir)

        self.set_data_reduction_params(data_obj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: stripe removal using Wavelet-FFT approach doi: 10.1364/OE.17.008567
   :platform: Unix
   :synopsis: A plugin removes ring artefacts

.. moduleauthor: Nghia Vo.
   Adapted from  tomopy source code:
   http://tomopy.github.io/tomopy/_modules/tomopy/algorithms/preprocess/stripe_removal.html
"""

import logging
import numpy as np
import pyfftw.interfaces.numpy_fft as fft
import pywt

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin

from savu.plugins.utils import register_plugin


@register_plugin
class RingRemovalWaveletfft(BaseFilter, CpuPlugin):
    """
    Ring artefact removal method

    :param nvalue: Order of the the Daubechies (DB) wavelets. Default: 5
    :param sigma: Damping parameter. Larger is stronger. Default: 1.
    :param level: Wavelet decomposition level. Default: 3.
    :param padFT: Padding for Fourier transform. Default: 20.
    """

    def __init__(self):
        logging.debug("Starting ring remogal using Wavelet-FFT approach")
        super(RingRemovalWaveletfft, self).__init__("RingRemovalWaveletfft")
        self.count = 0

    def set_filter_padding(self, in_data, out_data):
        self.pad = self.parameters['padFT']
        # don't currently have functionality to pad top/bottom but not
        # right/left so padding everywhere for now
        in_data[0].padding = {'pad_frame_edges': self.pad}
        out_data[0].padding = {'pad_frame_edges': self.pad}

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()[0]
        self.slice_dir = in_pData.get_slice_dimension()
        nDims = len(in_pData.get_shape())
        self.sslice = [slice(None)]*nDims
        sino_shape = list(in_pData.get_shape())
        if len(sino_shape) is 3:
            del sino_shape[self.slice_dir]

        self.width1 = sino_shape[1] + 2*self.pad
        self.height1 = sino_shape[0] + 2*self.pad

        n = np.abs(self.parameters['nvalue'])
        self.sigma = np.abs(self.parameters['sigma'])
        self.level = np.abs(self.parameters['level'])
        self.waveletname = 'db'+str(n)

    def process_frames(self, data):
        output = np.empty_like(data[0])
        nSlices = data[0].shape[self.slice_dir]
        for i in range(nSlices):
            self.sslice[self.slice_dir] = i
            sino = data[0][tuple(self.sslice)]
             # Wavelet decomposition.
            cH = []
            cV = []
            cD = []
            for j in range(self.level):
                sino, (cHt, cVt, cDt) = pywt.dwt2(sino, self.waveletname)
                cH.append(cHt)
                cV.append(cVt)
                cD.append(cDt)
             # FFT transform of horizontal frequency bands.
            for j in range(self.level):
                 # FFT
                fcV = np.fft.fftshift(np.fft.fft2(cV[j]))
                my, mx = fcV.shape
                # Damping of ring artifact information.
                y_hat = (np.arange(-my, my, 2, dtype='float') + 1) / 2.0
                damp = 1 - np.exp(-np.power(y_hat, 2) / (2 * np.power(self.sigma, 2)))
                fcV = np.multiply(fcV, np.transpose(np.tile(damp, (mx, 1))))

                 # Inverse FFT.
                cV[j] = np.real(np.fft.ifft2(np.fft.ifftshift(fcV)))
            # Wavelet reconstruction.
            for j in range(self.level)[::-1]:
                sino = sino[0:cH[j].shape[0], 0:cH[j].shape[1]]
                sino = pywt.idwt2((sino, (cH[j], cV[j], cD[j])), self.waveletname)
            if self.height1%2!=0:
                sino = sino[0:-1,:]
            if self.width1%2!=0:
                sino = sino[:,0:-1]
            output[self.sslice] = sino
        return output

    def get_plugin_pattern(self):
        return 'SINOGRAM'

    def get_max_frames(self):
        return 'multiple'




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomo_recon
   :platform: Unix
   :synopsis: runner for tests using the MPI framework

.. moduleauthor:: Nicola Wadespm <scientificsoftware@diamond.ac.uk>

"""

import unittest
from savu.test import test_utils as tu

from savu.test.travis.framework_tests.plugin_runner_test \
    import run_protected_plugin_runner
from savu.test.travis.framework_tests.plugin_runner_test \
    import run_protected_plugin_runner_no_process_list


class MultipleDatasetsTest(unittest.TestCase):

    def test_mm(self):
        data_file = tu.get_test_data_path('mm.nxs')
        process_file = tu.get_test_process_path('multiple_mm_inputs_test.nxs')
        run_protected_plugin_runner(tu.set_options(data_file,
                                                   process_file=process_file))

    def test_tomo1(self):
        options = tu.set_experiment('tomo')
        plugin = 'savu.plugins.basic_operations.basic_operations'
        data_dict = {'in_datasets': ['tomo', 'tomo'], 'out_datasets': ['test'],
                     'operations': ['tomo + tomo'],  'pattern': 'PROJECTION'}
        saver_dict = {}
        all_dicts = [{}, data_dict, saver_dict]
        run_protected_plugin_runner_no_process_list(options, plugin,
                                                    data=all_dicts)

    def test_tomo2(self):
        options = tu.set_experiment('tomo')
        plugin = 'savu.plugins.basic_operations.basic_operations'
        preview = ['10:-1:1:1', '10:-1:1:1', '10:-1:1:1']
        loader_dict = {'preview': preview}
        data_dict = {'in_datasets': ['tomo', 'tomo'], 'out_datasets': ['test'],
                     'operations': ['tomo + tomo'],  'pattern': 'PROJECTION'}
        all_dicts = [loader_dict, data_dict, {}]
        exp = run_protected_plugin_runner_no_process_list(options, plugin,
                                                          data=all_dicts)
        self.assertEqual(exp.index['in_data']['test'].get_shape(),
                         (81, 125, 150))

    def test_tomo3(self):
        options = tu.set_experiment('tomo')
        plugin = 'savu.plugins.basic_operations.basic_operations'
        preview = ['10:-1:10:1', '10:-1:10:1', '10:-1:10:1']
        loader_dict = {'preview': preview}
        data_dict = {'in_datasets': ['tomo', 'tomo'], 'out_datasets': ['test'],
                     'operations': ['tomo + tomo'],  'pattern': 'PROJECTION'}
        all_dicts = [loader_dict, data_dict, {}]
        exp = run_protected_plugin_runner_no_process_list(options, plugin,
                                                          data=all_dicts)
        exp = self.assertEqual(exp.index['in_data']['test'].get_shape(),
                               (9, 13, 15))

    def test_tomo4(self):
        options = tu.set_experiment('tomo')
        plugin = 'savu.plugins.basic_operations.basic_operations'
        preview = ['10:-10:10:1', '10:-10:10:1', '10:-10:10:1']
        loader_dict = {'preview': preview}
        data_dict = {'in_datasets': ['tomo', 'tomo'], 'out_datasets': ['test'],
                     'operations': ['tomo + tomo'],  'pattern': 'PROJECTION'}
        all_dicts = [loader_dict, data_dict, {}]
        exp = run_protected_plugin_runner_no_process_list(options, plugin,
                                                          data=all_dicts)
        exp = self.assertEqual(exp.index['in_data']['test'].get_shape(),
                               (8, 12, 15))

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: p2r_fly_scan_detector_loader
   :platform: Unix
   :synopsis: A class for loading p2r fly scan detector data.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import h5py
import logging
import numpy as np

from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.utils import register_plugin
import savu.core.utils as cu


@register_plugin
class P2rFlyScanDetectorLoader(BaseLoader):
    """
    A class to load p2r fly scan detector data from a Nexus file.

    :param data_path: Path to the data inside the \
        file. Default: 'entry1/tomo_entry/data/data'.
    :param image_key_path: Path to the image key entry inside the nxs \
        file. Default: 'entry1/tomo_entry/instrument/detector/image_key'.
    :param dark: Optional path to the dark field data file, nxs path and \
        scale value. Default: [None, None, 1].
    :param flat: Optional Path to the flat field data file, nxs path and \
        scale value. Default: [None, None, 1].
    :param angles: A python statement to be evaluated or a file. Default: None.
    :param 3d_to_4d: Set to true if this reshape is required. Default: False.
    :param ignore_flats: List of batch numbers of flats (start at 1) to \
        ignore. Default: None.
    """

    def __init__(self, name='P2rFlyScanDetectorLoader'):
        super(P2rFlyScanDetectorLoader, self).__init__(name)

    def setup(self):
        exp = self.exp

        data_obj = exp.create_data_object('in_data', 'tomo')

        data_obj.backing_file = \
            h5py.File(self.exp.meta_data.get("data_file"), 'r')

        data_obj.data = data_obj.backing_file[self.parameters['data_path']]

        self.__set_dark_and_flat(data_obj)

        if self.parameters['3d_to_4d']:
            if not self.parameters['angles']:
                raise Exception('Angles are required in the loader.')
            self.__setup_4d(data_obj)
            n_angles = self.__set_rotation_angles(data_obj)
            shape = self.__setup_3d_to_4d(data_obj, n_angles)
        else:
            if len(data_obj.data.shape) is 3:
                shape = self._setup_3d(data_obj)
            else:
                shape = self.__setup_4d(data_obj)
            self.__set_rotation_angles(data_obj)

        try:
            control = data_obj.backing_file['entry1/tomo_entry/control/data']
            data_obj.meta_data.set("control", control[...])
        except:
            logging.warning("No Control information available")

        nAngles = len(data_obj.meta_data.get('rotation_angle'))
        self.__check_angles(data_obj, nAngles)
        data_obj.set_original_shape(shape)
        self.set_data_reduction_params(data_obj)
        data_obj.data._set_dark_and_flat()

    def _setup_3d(self, data_obj):
        logging.debug("Setting up 3d tomography data.")
        rot = 0
        detY = 1
        detX = 2
        data_obj.set_axis_labels('rotation_angle.degrees',
                                 'detector_y.pixel',
                                 'detector_x.pixel')

        data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                             slice_dims=(rot,))
        data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                             slice_dims=(detY,))
        return data_obj.data.shape

    def __setup_3d_to_4d(self, data_obj, n_angles):
        logging.debug("setting up 4d tomography data from 3d input.")
        from savu.data.data_structures.data_type import Map_3dto4d_h5
        data_obj.data = Map_3dto4d_h5(data_obj.data, n_angles)
        return data_obj.data.get_shape()

    def __setup_4d(self, data_obj):
        logging.debug("setting up 4d tomography data.")
        rot = 0
        detY = 1
        detX = 2
        scan = 3

        data_obj.set_axis_labels('rotation_angle.degrees', 'detector_y.pixel',
                                 'detector_x.pixel', 'scan.number')

        data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                             slice_dims=(rot, scan))
        data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                             slice_dims=(detY, scan))
        return data_obj.data.shape

    def __set_dark_and_flat(self, data_obj):
        flat = self.parameters['flat'][0]
        dark = self.parameters['dark'][0]

        if not flat and not dark:
            self.__find_dark_and_flat(data_obj)
        else:
            self.__set_separate_dark_and_flat(data_obj)

    def __find_dark_and_flat(self, data_obj):
        ignore = self.parameters['ignore_flats'] if \
            self.parameters['ignore_flats'] else None
        try:
            image_key = \
                data_obj.backing_file[self.parameters['image_key_path']][...]

            from savu.data.data_structures.data_type import ImageKey
            data_obj.data = \
                ImageKey(data_obj, image_key, 0, ignore=ignore)
            #data_obj.set_shape(data_obj.data.get_shape())
        except KeyError:
            cu.user_message("An image key was not found.")
            try:
                from savu.data.data_structures.data_type import NoImageKey
                data_obj.data = NoImageKey(data_obj, None, 0)
                entry = 'entry1/tomo_entry/instrument/detector/'
                data_obj.data._set_flat_path(entry + 'flatfield')
                data_obj.data._set_dark_path(entry + 'darkfield')
            except KeyError:
                cu.user_message("Dark/flat data was not found in input file.")

    def __set_separate_dark_and_flat(self, data_obj):
        from savu.data.data_structures.data_type import NoImageKey
        try:
            image_key = \
                data_obj.backing_file[self.parameters['image_key_path']][...]
        except:
            image_key = None
        data_obj.data = NoImageKey(data_obj, image_key, 0)
        self.__set_data(data_obj, 'flat', data_obj.data._set_flat_path)
        self.__set_data(data_obj, 'dark', data_obj.data._set_dark_path)

    def __set_data(self, data_obj, name, func):
        path, entry, scale = self.parameters[name]

        if path.split('/')[0] == 'test_data':
            import os
            path = \
                os.path.dirname(os.path.abspath(__file__))+'/../../../' + path

        ffile = h5py.File(path, 'r')
        try:
            image_key = ffile[self.parameters['image_key_path']][...]
            func(ffile[entry], imagekey=image_key)
        except:
            func(ffile[entry])

        data_obj.data._set_scale(name, scale)

    def __set_rotation_angles(self, data_obj):
        angles = self.parameters['angles']
        if angles is None:
            try:
                entry = 'entry1/tomo_entry/data/rotation_angle'
                angles = data_obj.backing_file[entry][
                    (data_obj.data.get_image_key()) == 0, ...]
            except KeyError:
                logging.warning("No rotation angle entry found in input file.")
                angles = np.linspace(0, 180, data_obj.get_shape()[0])
        else:
            try:
                angles = eval(angles)
            except:
                try:
                    angles = np.loadtxt(angles)
                except:
                    raise Exception('Cannot set angles in loader.')

        data_obj.meta_data.set("rotation_angle", angles)
        return len(angles)

    def __check_angles(self, data_obj, n_angles):
        data_angles = data_obj.data.get_shape()[0]
        if data_angles != n_angles:
            raise Exception("The number of angles %s does not match the data "
                            "dimension length %s", n_angles, data_angles)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_datasets
   :platform: Unix
   :synopsis: Base class of plugin containing all dataset related functions

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import copy
import numpy as np

import savu.plugins.plugin_datasets_notes as notes
from savu.core.utils import docstring_parameter
from savu.data.data_structures.plugin_data import PluginData


class PluginDatasets(object):
    """
    The base class from which all plugins should inherit.
    """

    def __init__(self, *args, **kwargs):
        super(PluginDatasets, self).__init__()
        self.exp = None
        self.data_objs = {}
        self.variable_data_flag = False
        self.multi_params_dict = {}
        self.extra_dims = []
        self._max_itemsize = 0

    def __get_data_objects(self, dtype):
        """ Get the data objects associated with the plugin from the experiment
        data index.

        :params str dtype: "in_data" or "out_data"
        :returns: A list of data objects with the names given in
            ``self.parameters``
        :rtype: list(Data)
        """
        data_list = self.parameters[dtype + 'sets']
        data_objs = []
        for data in data_list:
            data_obj = self.exp.index[dtype][data]
            data_objs.append(data_obj)
        return data_objs

    def _clone_datasets(self):
        for data_obj in list(self.exp.index['out_data'].values()):
            if data_obj.raw and data_obj.data:
                data_obj.raw.create_next_instance(data_obj)
#                data_obj.clone = True

    def _finalise_datasets(self):
        in_data, out_data = self.get_datasets()
        for data in in_data + out_data:
            data._finalise_patterns()

    def _finalise_plugin_datasets(self):
        if 'dawn_runner' in list(self.exp.meta_data.get_dictionary().keys()):
            return

        in_pData, out_pData = self.get_plugin_datasets()

        params = {}
        for pData in in_pData + out_pData:
            pData._set_meta_data()
            params[pData] = pData._get_plugin_data_size_params()
        
        max_bytes = 0
        for key, value in params.items():
            if value['transfer_bytes'] > max_bytes:
                max_data = key
                max_bytes = value['transfer_bytes']

        # set mft and mfp for the largest dataset
        max_data.plugin_data_transfer_setup()
        to_set = list(set(params.keys()).difference(set([max_data])))

        for pData in to_set:
            if params[pData]['total_frames'] == params[max_data]['total_frames']:
                pData.plugin_data_transfer_setup(copy=max_data)
            else:
                if pData.max_frames == 'multiple':
                    msg = "If a plugin reduces the number of frames, the " \
                        "number of frames cannot be 'multiple'."
                    raise Exception(msg)
                pData.plugin_data_transfer_setup(calc=max_data)

    def __set_in_datasets(self):
        """ Set the in_data objects.

        :returns: the in_datasets associated with the plugin.
        :rtype: list[Data]
        """
        return self.__get_data_objects('in_data')

    def __set_out_datasets(self):
        """ Set the out_data objects.

        If the out_datasets do not exist inside the experiment then create
        them.

        :returns: the out_datasets associated with the plugin.
        :rtype: list[Data]
        """
        try:
            out_data = self.__get_data_objects('out_data')
        except KeyError:
            out_data = []
            for data in self.parameters['out_datasets']:
                self.exp.create_data_object("out_data", data)
            out_data = self.__get_data_objects('out_data')
        for data in out_data:
            data.extra_dims = self.extra_dims
        return out_data

    def _get_plugin_data(self, data_list):
        """ Encapsulate a PluginData object in each dataset associated with
        the plugin.

        :params list(Data) data_list: A list of Data objects used in a plugin.
        :returns: A list of PluginData objects.
        :rtype: list(PluginData)
        """
        pData_list = []
        used = set()
        unique_data_list = \
            [x for x in data_list if x not in used and (used.add(x) or True)]
        for data in unique_data_list:
            pData_list.append(PluginData(data, self))
            pData_list[-1].extra_dims = self.extra_dims
            pData_list[-1].multi_params_dict = self.multi_params_dict
        return pData_list

    def _set_plugin_dataset_names(self):
        """ Fill in empty in/out_dataset entries with default values.
        """
        params = self.parameters
        orig_in = copy.copy(params['in_datasets'])
        in_names = self._set_in_dataset_names(params)
        # case that an extra in_dataset is added in the plugin
        in_names = orig_in if len(orig_in) and \
            len(in_names) > len(orig_in) else in_names        
        self._set_out_dataset_names(params, in_names)
        # update the entry in the process list
        data_dict = {'in_datasets': params['in_datasets'],
                      'out_datasets': params['out_datasets']}
        idx = self.exp.meta_data.get('nPlugin')
        self.exp.meta_data.plugin_list._update_datasets(idx, data_dict)

    def _set_in_dataset_names(self, params):
        dIn = params['in_datasets']
        dIn = dIn if isinstance(dIn, list) else [dIn]
        dIn = self.exp._set_all_datasets('in_data') if len(dIn) ==0 else dIn
        params['in_datasets'] = dIn
        nIn = self.nInput_datasets() # datasets many be added dynamically here
        return self.check_nDatasets(params['in_datasets'], nIn, 'in_data')

    def _set_out_dataset_names(self, params, in_names):
        dOut = params['out_datasets'] if 'out_datasets' in params.keys() else []
        dOut = dOut if isinstance(dOut, list) else [dOut]
        dOut = (copy.copy(in_names) if len(dOut) == 0 else dOut)
        clones = self.nClone_datasets()
        params['out_datasets'] = dOut
        nOut = self.nOutput_datasets()
        names = self.check_nDatasets(params['out_datasets'], nOut,
                                     "out_data", clones=clones)
        if clones:
            dOut.extend(['itr_clone' + str(i) for i in range(clones)])
    
        for i in range(len(names)):
            new = names[i].split('in_datasets')
            if len(new) == 2:
                names[i] = in_names[int(list(new[1])[1])]
        params["out_datasets"] = names
        return names

    def check_nDatasets(self, names, nSets, dtype, clones=0):
        nSets = len(self.parameters[dtype + 'sets']) if nSets=='var' else nSets
        if len(names) is not (nSets - clones):
            if nSets == 0:
                names = []
            else:
                msg = "ERROR: Broken plugin chain. \n Please name the %s %s " \
                "sets associated with the plugin %s in the process file." % \
                (str(nSets), dtype, self.name)
                raise Exception(msg)
        return names

    def _set_plugin_datasets(self):
        """ Populate ``self.parameters`` in/out_datasets and
        plugin_in/out_datasets with the relevant objects (Data or PluginData).
        """
        if not self.exp._get_dataset_names_complete():
            self._set_plugin_dataset_names()
        self.parameters['in_datasets'] = self.__set_in_datasets()
        self.parameters['out_datasets'] = self.__set_out_datasets()
        self.parameters['plugin_in_datasets'] = \
            self._get_plugin_data(self.parameters['in_datasets'])
        self.parameters['plugin_out_datasets'] = \
            self._get_plugin_data(self.parameters['out_datasets'])

    @docstring_parameter('PluginData', 'in')
    @docstring_parameter(notes.datasets_notes.__doc__)
    def get_plugin_in_datasets(self):
        """ {0} """
        return self.parameters['plugin_in_datasets']

    @docstring_parameter('PluginData', 'out')
    @docstring_parameter(notes.datasets_notes.__doc__)
    def get_plugin_out_datasets(self):
        """ {0} """
        return self.parameters['plugin_out_datasets']

    @docstring_parameter("PluginData")
    @docstring_parameter(notes.two_datasets_notes.__doc__)
    def get_plugin_datasets(self):
        """ {0} """
        return self.get_plugin_in_datasets(), self.get_plugin_out_datasets()

    @docstring_parameter("Data", "in")
    @docstring_parameter(notes.datasets_notes.__doc__)
    def get_in_datasets(self):
        """ {0} """
        return self.parameters['in_datasets']

    @docstring_parameter("Data", "out")
    @docstring_parameter(notes.datasets_notes.__doc__)
    def get_out_datasets(self):
        """ {0} """
        return self.parameters['out_datasets']

    @docstring_parameter("PluginData")
    @docstring_parameter(notes.two_datasets_notes.__doc__)
    def get_datasets(self):
        """ {0} """
        return self.get_in_datasets(), self.get_out_datasets()

    @docstring_parameter("in")
    @docstring_parameter(notes.mData_notes.__doc__)
    def get_in_meta_data(self):
        """ {0} """
        return self.__set_meta_data(self.parameters['in_datasets'], 'in_data')

    @docstring_parameter("out")
    @docstring_parameter(notes.mData_notes.__doc__)
    def get_out_meta_data(self):
        """ {0} """
        return self.__set_meta_data(self.parameters['out_datasets'],
                                    'out_data')

    def get(self):
        """ Get a list of meta_data objects associated with the
        in/out_datasets.

        :returns: All MetaData objects associated with out data objects.
        :rtype: list(MetaData(in_datasets)), list(MetaData(out_datasets))
        """
        return self.get_in_meta_data(), self.get_out_meta_data()

    def __set_meta_data(self, data_list, dtype):
        """ Append all MetaData objs associated with specified datasets to a
        list.

        :params list(Data) data_list:
        :returns: All MetaData objects associated with data objects in
            data_list
        :rtype: list(MetaData)
        """
        meta_data = []
        for data in data_list:
            meta_data.append(data.meta_data)
        return meta_data

    def _set_unknown_shape(self, data, key):
        try:
            return (len(data.meta_data.get(key)),)
        except KeyError:
            return (0,)



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 29 15:42:30 2019

@author: qmm55171
"""

import argparse

from scripts.config_generator.content import Content


def arg_parser():
    """ Arg parser for command line arguments.
    """
    parser = argparse.ArgumentParser(prog='update_template')
    parser.add_argument('nxsfile', help='The nexus file')
    parser.add_argument('plugin', help='Plugin name')
    parser.add_argument('param', help='plugin parameter name')
    return parser.parse_args()

def get_parameter_value(nxsfile, plugin_name, plugin_param):
    content = Content()
    content.fopen(nxsfile)
    plist = content.plugin_list.plugin_list
    
    for plugin in plist:
        if plugin['name'] == plugin_name:
            if plugin_param in list(plugin['data'].keys()):
                print(plugin_name, plugin['data'][plugin_param])

if __name__ == '__main__':
    arg = arg_parser()
    get_parameter_value(arg.nxsfile, arg.plugin, arg.param)
